# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""Test PlotBuilder.curve parameters"""

import numpy as np
import pytest
from guidata.qthelpers import exec_dialog, qt_app_context
from qtpy.QtCore import Qt
from qwt import QwtPlotCurve

from plotpy.builder import make


def show_items_qtbot(items, type="curve"):
    """Plot curve in a dialog"""
    win = make.dialog(type=type)
    plot = win.manager.get_plot()
    for item in items:
        plot.add_item(item)
    exec_dialog(win)


_COLOR_TO_HEX = {"red": "#ff0000", "blue": "#0000ff"}


def _make_curve_style(shade, curvestyle, baseline):
    """Make curve with curve style parameters"""
    x = np.linspace(-10, 10, 200)
    y = np.sin(x)
    curve = make.curve(x, y, shade=shade, curvestyle=curvestyle, baseline=baseline)
    assert curve.baseline() == baseline
    brush = curve.brush()
    assert brush.color().alphaF() == pytest.approx(shade)
    assert curve.style() == getattr(QwtPlotCurve, curvestyle)
    return curve


@pytest.mark.parametrize("shade", [0])
@pytest.mark.parametrize("curvestyle", ["Lines", "Sticks", "Steps", "Dots", "NoCurve"])
@pytest.mark.parametrize("baseline", [0.0])
def test_builder_curve_curve_style(shade, curvestyle, baseline):
    """Test curve parameters of curve() method"""
    with qt_app_context(exec_loop=False):
        curve = _make_curve_style(shade, curvestyle, baseline)
        show_items_qtbot([curve], "curve")


@pytest.mark.parametrize("shade", [0, 0.4, 1.0])
@pytest.mark.parametrize("curvestyle", ["Lines"])
@pytest.mark.parametrize("baseline", [0.0, 1.0])
def test_builder_curve_curve_shade_baseline(shade, curvestyle, baseline):
    """Test curve parameters of curve() method"""
    with qt_app_context(exec_loop=False):
        curve = _make_curve_style(shade, curvestyle, baseline)
        show_items_qtbot([curve], "curve")


def _make_curve_dsamp(dsamp_factor, use_dsamp):
    """Make curve with downsampling parameters"""
    x = np.linspace(-10, 10, 1000)
    y = np.sin(x)
    curve = make.curve(x, y, dsamp_factor=dsamp_factor, use_dsamp=use_dsamp)
    if use_dsamp:
        assert curve.dataSize() == x[::dsamp_factor].size
    return curve


@pytest.mark.parametrize("dsamp_factor", [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 15, 30, 50])
@pytest.mark.parametrize("use_dsamp", [True])
def test_builder_curve_dsamp_on(dsamp_factor, use_dsamp):
    """Test downsampling parameters of curve() method: use_dsamp=True"""
    with qt_app_context(exec_loop=False):
        curve = _make_curve_dsamp(dsamp_factor, use_dsamp)
        show_items_qtbot([curve], "curve")


@pytest.mark.parametrize("dsamp_factor", [1, 2])
@pytest.mark.parametrize("use_dsamp", [False])
def test_builder_curve_dsamp_off(dsamp_factor, use_dsamp):
    """Test downsampling parameters of curve() method: use_dsamp=False"""
    with qt_app_context(exec_loop=False):
        curve = _make_curve_dsamp(dsamp_factor, use_dsamp)
        show_items_qtbot([curve], "curve")


def _make_curve_linestyle(color, linestyle, linewidth):
    """Make curve with line parameters"""
    x = np.linspace(-10, 10, 200)
    y = np.sin(x)
    curve = make.curve(x, y, color=color, linestyle=linestyle, linewidth=linewidth)
    pen = curve.pen()
    assert pen.color().name() == _COLOR_TO_HEX[color]
    assert pen.style() == getattr(Qt, linestyle)
    assert pen.width() == linewidth
    return curve


@pytest.mark.parametrize("color", ["red"])
@pytest.mark.parametrize(
    "linestyle",
    ["SolidLine", "DashLine", "DotLine", "DashDotLine", "DashDotDotLine", "NoPen"],
)
@pytest.mark.parametrize("linewidth", [1, 2])
def test_builder_curve_line_style(color, linestyle, linewidth):
    """Test line parameters of curve() method"""
    with qt_app_context(exec_loop=False):
        curve = _make_curve_linestyle(color, linestyle, linewidth)
        show_items_qtbot([curve], "curve")


@pytest.mark.parametrize("color", ["red", "blue"])
@pytest.mark.parametrize("linestyle", ["SolidLine"])
@pytest.mark.parametrize("linewidth", [1, 2])
def test_builder_curve_line_color(color, linestyle, linewidth):
    """Test line parameters of curve() method"""
    with qt_app_context(exec_loop=False):
        curve = _make_curve_linestyle(color, linestyle, linewidth)
        show_items_qtbot([curve], "curve")


def _make_curve_marker(marker, markersize, markerfacecolor, markeredgecolor):
    """Make curve with marker parameters"""
    x = np.linspace(-10, 10, 200)
    y = np.sin(x)
    curve = make.curve(
        x,
        y,
        marker=marker,
        markersize=markersize,
        markerfacecolor=markerfacecolor,
        markeredgecolor=markeredgecolor,
    )
    symbol = curve.symbol()
    assert symbol.style() == getattr(type(symbol), marker)
    assert symbol.size().width() == markersize
    assert symbol.size().height() == markersize
    assert symbol.brush().color().name() == _COLOR_TO_HEX[markerfacecolor]
    assert symbol.pen().color().name() == _COLOR_TO_HEX[markeredgecolor]
    return curve


@pytest.mark.parametrize(
    "marker",
    [
        "Cross",
        "Ellipse",
        "Star1",
        "XCross",
        "Rect",
        "Diamond",
        "UTriangle",
        "DTriangle",
        "RTriangle",
        "LTriangle",
        "Star2",
        "NoSymbol",
    ],
)
@pytest.mark.parametrize("markersize", [5])
@pytest.mark.parametrize("markerfacecolor", ["red"])
@pytest.mark.parametrize("markeredgecolor", ["blue"])
def test_builder_curve_marker_params_symbol(
    marker, markersize, markerfacecolor, markeredgecolor
):
    """Test marker parameters of curve() methodg"""
    with qt_app_context(exec_loop=False):
        curve = _make_curve_marker(marker, markersize, markerfacecolor, markeredgecolor)
        show_items_qtbot([curve], "curve")


@pytest.mark.parametrize("marker", ["Cross"])
@pytest.mark.parametrize("markersize", [1, 5, 10])
@pytest.mark.parametrize("markerfacecolor", ["red", "blue"])
@pytest.mark.parametrize("markeredgecolor", ["red", "blue"])
def test_builder_curve_marker_size_color(
    marker, markersize, markerfacecolor, markeredgecolor
):
    """Test marker parameters of curve() methodg"""
    with qt_app_context(exec_loop=False):
        curve = _make_curve_marker(marker, markersize, markerfacecolor, markeredgecolor)
        show_items_qtbot([curve], "curve")



# cython: language_level=3
# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/__init__.py for details)

"""Mandelbrot algorithm"""

cimport cython
cimport numpy as np

@cython.profile(False)
cdef inline int mandel(double real, double imag, int nmax):
    cdef double z_real=0., z_imag=0.
    cdef int i
    for i in range(nmax):
        z_real, z_imag = (z_real*z_real - z_imag*z_imag + real,
                          2*z_real*z_imag + imag)
        if z_real*z_real + z_imag*z_imag > 4:
            return i
    return -1

@cython.boundscheck(False)
def mandelbrot(double x1, double y1, double x2, double y2,
               data, unsigned int nmax):
    """Compute Mandelbrot fractal"""
    cdef double dx, dy, real, imag
    cdef unsigned int row, col
    cdef unsigned int rows = data.shape[0]
    cdef unsigned int cols = data.shape[1]

    dx = (x2-x1)/(cols-1)
    dy = (y2-y1)/(rows-1)

    for col in range(cols):
        real = x1 + col*dx
        for row in range(rows):
            imag = y1 + row*dy
            data[row, col] = mandel(real, imag, nmax)



import unittest

import numpy as np

from plotpy.items.contour import compute_contours
from plotpy.tests.data import gen_xyz_data


class ContourTest(unittest.TestCase):
    """
    Test case for contour() function

    Test data are borrowed from https://matplotlib.org/gallery/images_contours_and_fields/contour_demo.html#sphx-glr-gallery-images-contours-and-fields-contour-demo-py

    """

    def setUp(self):
        self.X, self.Y, self.Z = gen_xyz_data()

    def test_contour_level_1(self):
        """Test that contour() returns one closed line when level is 1.0"""
        # Ortho coord
        lines = compute_contours(self.Z, 1.0, self.X, self.Y)
        assert len(lines) == 1
        assert np.all(np.equal(lines[0].vertices[0], lines[0].vertices[-1]))
        assert lines[0].level == 1.0

        # Grid coord
        lines = compute_contours(self.Z, 1.0, self.X, self.Y)
        assert len(lines) == 1
        assert np.all(np.equal(lines[0].vertices[0], lines[0].vertices[-1]))
        assert lines[0].level == 1.0

    def test_contour_level_diag_0(self):
        """Test that contour() returns opened lines (diagonal) when level is 0.0"""
        lines = compute_contours(self.Z, 0.0, self.X, self.Y)
        assert len(lines) == 1
        assert np.any(lines[0].vertices[0] != lines[0].vertices[-1])
        assert lines[0].level == 0.0

    def test_contour_level_opened_neg_0_5(self):
        """Test that contour() returns opened lines when level is -0.5"""
        lines = compute_contours(self.Z, -0.5, self.X, self.Y)
        assert len(lines) == 1
        assert np.any(lines[0].vertices[0] != lines[0].vertices[-1])
        assert lines[0].level == -0.5

    def test_contour_abs_level_1(self):
        """Test that contour() returns two closed lines when level is 1.0
        and data are absolute.
        """
        lines = compute_contours(np.abs(self.Z), 1.0, self.X, self.Y)
        assert len(lines) == 2
        for line in lines:
            assert np.all(np.equal(line.vertices[0], line.vertices[-1]))
            assert line.level == 1.0



# -*- coding: utf-8 -*-

from __future__ import annotations

import weakref
from typing import TYPE_CHECKING, Any

from guidata.qthelpers import create_action
from guidata.utils.misc import assert_interfaces_valid
from qtpy import QtCore as QC
from qtpy import QtWidgets as QW

from plotpy.constants import ID_CONTRAST, ID_ITEMLIST, ID_XCS, ID_YCS
from plotpy.interfaces import IPlotManager
from plotpy.plot import BasePlot
from plotpy.tools import (
    AboutTool,
    AnnotatedCircleTool,
    AnnotatedEllipseTool,
    AnnotatedObliqueRectangleTool,
    AnnotatedPointTool,
    AnnotatedRectangleTool,
    AnnotatedSegmentTool,
    AntiAliasingTool,
    AspectRatioTool,
    AverageCrossSectionTool,
    AxisScaleTool,
    BasePlotMenuTool,
    ColormapTool,
    ContrastPanelTool,
    CopyToClipboardTool,
    CrossSectionTool,
    CurveStatsTool,
    DeleteItemTool,
    DisplayCoordsTool,
    DoAutoscaleTool,
    DownSamplingTool,
    DummySeparatorTool,
    EditItemDataTool,
    ExportItemDataTool,
    HelpTool,
    ImageStatsTool,
    ItemCenterTool,
    ItemListPanelTool,
    LabelTool,
    PrintTool,
    RectangularSelectionTool,
    RectZoomTool,
    ReverseColormapTool,
    ReverseXAxisTool,
    ReverseYAxisTool,
    SaveAsTool,
    SelectTool,
    SnapshotTool,
    XCSPanelTool,
    YCSPanelTool,
    ZAxisLogTool,
)

if TYPE_CHECKING:
    from typing import Callable

    from qtpy.QtCore import Qt
    from qtpy.QtGui import QIcon, QKeySequence
    from qwt import QwtPlotCanvas, QwtScaleDiv

    from plotpy.panels import (
        ContrastAdjustment,
        PanelWidget,
        PlotItemList,
        XCrossSection,
        YCrossSection,
    )
    from plotpy.tools.base import GuiTool, GuiToolT


class DefaultPlotID:
    pass


class PlotManager:
    """
    Construct a PlotManager object, a 'controller' that organizes relations
    between plots (:py:class:`.BasePlot`), panels, tools and toolbars

    Args:
        main (QWidget): The main parent widget
    """

    __implements__ = (IPlotManager,)

    def __init__(self, main: QW.QWidget) -> None:
        self.main = main  # The main parent widget
        self.plots = {}  # maps ids to instances of BasePlot
        self.panels = {}  # Qt widgets that need to know about the plots
        self.tools = []
        self.toolbars = {}
        self.active_tool = None
        self.default_tool = None
        self.default_plot = None
        self.default_toolbar = None
        self.synchronized_plots = {}
        self.groups = {}  # Action groups for grouping QActions
        # Keep track of the registration sequence (plots, panels, tools):
        self._first_tool_flag = True

    def add_plot(self, plot: BasePlot, plot_id: Any = DefaultPlotID) -> None:
        """
        Register a plot to the plot manager:
            * plot: :py:class:`.BasePlot`
            * plot_id (default id is the plot object's id: ``id(plot)``):
              unique ID identifying the plot (any Python object),
              this ID will be asked by the manager to access this plot later.

        Plot manager's registration sequence is the following:
            1. add plots
            2. add panels
            3. add tools
        """
        if plot_id is DefaultPlotID:
            plot_id = id(plot)
        assert plot_id not in self.plots
        assert isinstance(plot, BasePlot)
        assert not self.tools, "tools must be added after plots"
        assert not self.panels, "panels must be added after plots"
        self.plots[plot_id] = plot
        if len(self.plots) == 1:
            self.default_plot = plot
        plot.set_manager(self, plot_id)
        # Connecting signals
        plot.SIG_ITEMS_CHANGED.connect(self.update_tools_status)
        plot.SIG_ACTIVE_ITEM_CHANGED.connect(self.update_tools_status)
        plot.SIG_PLOT_AXIS_CHANGED.connect(self.plot_axis_changed)

    def set_default_plot(self, plot: BasePlot) -> None:
        """
        Set default plot

        The default plot is the plot on which tools and panels will act.
        """
        self.default_plot = plot

    def get_default_plot(self) -> BasePlot:
        """
        Return default plot

        The default plot is the plot on which tools and panels will act.
        """
        return self.default_plot

    def add_panel(self, panel: PanelWidget) -> None:
        """
        Register a panel to the plot manager

        Plot manager's registration sequence is the following:
            1. add plots
            2. add panels
            3. add tools
        """
        assert panel.PANEL_ID not in self.panels
        assert not self.tools, "tools must be added after panels"
        self.panels[panel.PANEL_ID] = panel
        panel.register_panel(self)

    def configure_panels(self) -> None:
        """
        Call all the registred panels 'configure_panel' methods to finalize the
        object construction (this allows to use tools registered to the same
        plot manager as the panel itself with breaking the registration
        sequence: "add plots, then panels, then tools")
        """
        for panel_id in self.panels:
            panel = self.get_panel(panel_id)
            panel.configure_panel()

    def add_toolbar(self, toolbar: QW.QToolBar, toolbar_id: str = "default") -> None:
        """
        Add toolbar to the plot manager
            toolbar: a QToolBar object
            toolbar_id: toolbar's id (default id is string "default")
        """
        assert toolbar_id not in self.toolbars
        self.toolbars[toolbar_id] = toolbar
        if self.default_toolbar is None:
            self.default_toolbar = toolbar

    def set_default_toolbar(self, toolbar: QW.QToolBar) -> None:
        """
        Set default toolbar
        """
        self.default_toolbar = toolbar

    def get_default_toolbar(self) -> QW.QToolBar:
        """
        Return default toolbar
        """
        return self.default_toolbar

    def add_tool(self, ToolKlass: type[GuiToolT], *args, **kwargs) -> GuiToolT:
        """
        Register a tool to the manager
            * ToolKlass: tool's class (see :ref:`tools`)
            * args: arguments sent to the tool's class
            * kwargs: keyword arguments sent to the tool's class

        Plot manager's registration sequence is the following:
            1. add plots
            2. add panels
            3. add tools
        """
        if self._first_tool_flag:
            # This is the very first tool to be added to this manager
            self._first_tool_flag = False
            self.configure_panels()
        tool = ToolKlass(self, *args, **kwargs)
        self.tools.append(tool)
        for plot in list(self.plots.values()):
            tool.register_plot(plot)
        if len(self.tools) == 1 or self.default_tool is None:
            self.default_tool = tool
        return tool

    def get_tool(self, ToolKlass: type[GuiToolT]) -> GuiToolT | None:
        """Return tool instance from its class

        Args:
            ToolKlass: tool's class (see :ref:`tools`)

        Returns:
            GuiTool: tool instance
        """
        for tool in self.tools:
            if isinstance(tool, ToolKlass):
                return tool

    def add_separator_tool(self, toolbar_id: str | None = None) -> None:
        """
        Register a separator tool to the plot manager: the separator tool is
        just a tool which insert a separator in the plot context menu

        Args:
            toolbar_id: toolbar's id (default to None)
        """
        if toolbar_id is None:
            for _id, toolbar in list(self.toolbars.items()):
                if toolbar is self.get_default_toolbar():
                    toolbar_id = _id
                    break
        self.add_tool(DummySeparatorTool, toolbar_id)

    def set_default_tool(self, tool: GuiTool) -> None:
        """
        Set default tool

        Args:
            tool: tool instance
        """
        self.default_tool = tool

    def get_default_tool(self) -> GuiTool:
        """
        Get default tool

        Returns:
            GuiTool: tool instance
        """
        return self.default_tool

    def activate_default_tool(self) -> None:
        """
        Activate default tool
        """
        self.get_default_tool().activate()

    def get_active_tool(self) -> GuiTool:
        """
        Return active tool

        Returns:
            GuiTool: tool instance
        """
        return self.active_tool

    def set_active_tool(self, tool: GuiTool | None = None) -> None:
        """
        Set active tool (if tool argument is None, the active tool will be
        the default tool)

        Args:
            tool: tool instance or None
        """
        self.active_tool = tool

    def get_plot(self, plot_id: Any = DefaultPlotID) -> BasePlot:
        """
        Return plot associated to `plot_id` (if method is called without
        specifying the `plot_id` parameter, return the default plot)

        Args:
            plot_id: plot's id (optional, default to DefaultPlotID)
        """
        if plot_id is DefaultPlotID:
            return self.default_plot
        return self.plots[plot_id]

    def get_plots(self) -> list[BasePlot]:
        """
        Return all registered plots

        Returns:
            list[BasePlot]: list of plots
        """
        return list(self.plots.values())

    def get_active_plot(self) -> BasePlot:
        """
        Return the active plot

        The active plot is the plot whose canvas has the focus
        otherwise it's the "default" plot

        Returns:
            BasePlot: plot instance
        """
        for plot in list(self.plots.values()):
            canvas: QwtPlotCanvas = plot.canvas()
            if canvas.hasFocus():
                return plot
        return self.default_plot

    def get_tool_group(self, groupname: str) -> QW.QActionGroup:
        """
        Return the QActionGroup associated to `groupname`

        Args:
            groupname: group's name

        Returns:
            QActionGroup: action group
        """
        group = self.groups.get(groupname, None)
        if group is None:
            group = QW.QActionGroup(self.main)
            self.groups[groupname] = weakref.ref(group)
            return group
        else:
            return group()

    def get_main(self) -> QW.QWidget:
        """
        Return the main (parent) widget

        Note that for py:class:`.plot.PlotWidget` objects, this method will
        return the widget itself because the plot manager is integrated to it.

        Returns:
            QWidget: main widget
        """
        return self.main

    def set_main(self, main: QW.QWidget) -> None:
        """
        Set the main (parent) widget

        Args:
            main: main widget
        """
        self.main = main

    def get_panel(self, panel_id: str) -> PanelWidget:
        """
        Return panel from its ID
        Panel IDs are listed in module plotpy.panels

        Args:
            panel_id: panel's id

        Returns:
            PanelWidget: panel widget
        """
        return self.panels.get(panel_id, None)

    def get_itemlist_panel(self) -> PlotItemList:
        """
        Convenience function to get the `item list panel`

        Return None if the item list panel has not been added to this manager

        Returns:
            PlotItemList: item list panel
        """
        return self.get_panel(ID_ITEMLIST)

    def get_contrast_panel(self) -> ContrastAdjustment:
        """
        Convenience function to get the `contrast adjustment panel`

        Return None if the contrast adjustment panel has not been added
        to this manager
        """
        return self.get_panel(ID_CONTRAST)

    def set_contrast_range(self, zmin: float, zmax: float) -> None:
        """
        Convenience function to set the `contrast adjustment panel` range

        This is strictly equivalent to the following::

            # Here, *widget* is for example a PlotWidget instance
            # (the same apply for PlotWidget or any
            #  class deriving from PlotManager)
            widget.get_contrast_panel().set_range(zmin, zmax)

        Args:
            zmin: minimum value
            zmax: maximum value
        """
        self.get_contrast_panel().set_range(zmin, zmax)

    def get_xcs_panel(self) -> XCrossSection:
        """
        Convenience function to get the `X-axis cross section panel`

        Return None if the X-axis cross section panel has not been added
        to this manager

        Returns:
            XCrossSection: X-axis cross section panel
        """
        return self.get_panel(ID_XCS)

    def get_ycs_panel(self) -> YCrossSection:
        """
        Convenience function to get the `Y-axis cross section panel`

        Return None if the Y-axis cross section panel has not been added
        to this manager

        Returns:
            YCrossSection: Y-axis cross section panel
        """
        return self.get_panel(ID_YCS)

    def update_cross_sections(self) -> None:
        """
        Convenience function to update the `cross section panels` at once

        This is strictly equivalent to the following::

            # Here, *widget* is for example a PlotWidget instance
            # (the same apply for any other class deriving from PlotManager)
            widget.get_xcs_panel().update_plot()
            widget.get_ycs_panel().update_plot()
        """
        self.get_xcs_panel().update_plot()
        self.get_ycs_panel().update_plot()

    def get_toolbar(self, toolbar_id: str = "default") -> QW.QToolBar:
        """
        Return toolbar from its ID

        Args:
            toolbar_id: toolbar's id (default id is string "default")

        Returns:
            QToolBar: toolbar
        """
        return self.toolbars.get(toolbar_id, None)

    def get_context_menu(self, plot: BasePlot | None = None) -> QW.QMenu:
        """
        Return widget context menu -- built using active tools

        Args:
            plot: plot instance (default to None)

        Returns:
            QMenu: context menu
        """
        if plot is None:
            plot = self.get_plot()
        menu = QW.QMenu(plot)
        self.update_tools_status(plot)
        for tool in self.tools:
            tool.setup_context_menu(menu, plot)
        return menu

    def update_tools_status(self, plot: BasePlot | None = None) -> None:
        """
        Update tools for current plot

        Args:
            plot: plot instance (default to None)
        """
        if plot is None:
            plot = self.get_plot()
        for tool in self.tools:
            tool.update_status(plot)

    def create_action(
        self,
        title: str,
        triggered: Callable | None = None,
        toggled: Callable | None = None,
        shortcut: QKeySequence | None = None,
        icon: QIcon | None = None,
        tip: str | None = None,
        checkable: bool | None = None,
        context: Qt.ShortcutContext = QC.Qt.ShortcutContext.WindowShortcut,
        enabled: bool | None = None,
    ):
        """
        Create a new QAction

        Args:
            parent (QWidget or None): Parent widget
            title (str): Action title
            triggered (Callable or None): Triggered callback
            toggled (Callable or None): Toggled callback
            shortcut (QKeySequence or None): Shortcut
            icon (QIcon or None): Icon
            tip (str or None): Tooltip
            checkable (bool or None): Checkable
            context (Qt.ShortcutContext): Shortcut context
            enabled (bool or None): Enabled

        Returns:
            QAction: New action
        """
        return create_action(
            self.main,
            title,
            triggered=triggered,
            toggled=toggled,
            shortcut=shortcut,
            icon=icon,
            tip=tip,
            checkable=checkable,
            context=context,
            enabled=enabled,
        )

    # The following methods provide some sets of tools that
    # are often registered together
    def register_standard_tools(self) -> None:
        """
        Registering basic tools for standard plot dialog
        --> top of the context-menu
        """
        t = self.add_tool(SelectTool)
        self.set_default_tool(t)
        self.add_tool(RectangularSelectionTool, intersect=False)
        self.add_tool(RectZoomTool)
        self.add_tool(DoAutoscaleTool)
        self.add_tool(BasePlotMenuTool, "item")
        self.add_tool(ExportItemDataTool)
        self.add_tool(EditItemDataTool)
        self.add_tool(ItemCenterTool)
        self.add_tool(DeleteItemTool)
        self.add_separator_tool()
        self.add_tool(BasePlotMenuTool, "grid")
        self.add_tool(BasePlotMenuTool, "axes")
        self.add_tool(DisplayCoordsTool)
        if self.get_itemlist_panel():
            self.add_tool(ItemListPanelTool)

    def register_curve_tools(self) -> None:
        """
        Register only curve-related tools

        .. seealso::

            :py:meth:`.plot.manager.PlotManager.add_tool`

            :py:meth:`.plot.manager.PlotManager.register_standard_tools`

            :py:meth:`.plot.manager.PlotManager.register_other_tools`

            :py:meth:`.plot.manager.PlotManager.register_image_tools`

            :py:meth:`.plot.manager.PlotManager.register_all_tools`
        """
        self.add_tool(CurveStatsTool)
        self.add_tool(AntiAliasingTool)
        self.add_tool(AxisScaleTool)
        self.add_tool(DownSamplingTool)

    def register_image_tools(self) -> None:
        """
        Register only image-related tools

        .. seealso::

            :py:meth:`.plot.manager.PlotManager.add_tool`

            :py:meth:`.plot.manager.PlotManager.register_standard_tools`

            :py:meth:`.plot.manager.PlotManager.register_other_tools`

            :py:meth:`.plot.manager.PlotManager.register_curve_tools`

            :py:meth:`.plot.manager.PlotManager.register_all_tools`
        """
        self.add_tool(ColormapTool)
        self.add_tool(ReverseColormapTool)
        self.add_tool(ReverseXAxisTool)
        self.add_tool(ReverseYAxisTool)
        self.add_tool(ZAxisLogTool)
        self.add_tool(AspectRatioTool)
        if self.get_contrast_panel():
            self.add_tool(ContrastPanelTool)
        self.add_tool(SnapshotTool)
        self.add_tool(ImageStatsTool)
        if self.get_xcs_panel() and self.get_ycs_panel():
            self.add_tool(XCSPanelTool)
            self.add_tool(YCSPanelTool)
            self.add_tool(CrossSectionTool)
            self.add_tool(AverageCrossSectionTool)

    def register_other_tools(self) -> None:
        """
        Register other common tools

        .. seealso::

            :py:meth:`.plot.manager.PlotManager.add_tool`

            :py:meth:`.plot.manager.PlotManager.register_standard_tools`

            :py:meth:`.plot.manager.PlotManager.register_curve_tools`

            :py:meth:`.plot.manager.PlotManager.register_image_tools`

            :py:meth:`.plot.manager.PlotManager.register_all_tools`
        """
        self.add_tool(SaveAsTool)
        self.add_tool(CopyToClipboardTool)
        self.add_tool(PrintTool)
        self.add_tool(HelpTool)
        self.add_tool(AboutTool)

    def register_all_curve_tools(self) -> None:
        """
        Register standard, curve-related and other tools

        .. seealso::

            :py:meth:`.plot.manager.PlotManager.add_tool`

            :py:meth:`.plot.manager.PlotManager.register_standard_tools`

            :py:meth:`.plot.manager.PlotManager.register_other_tools`

            :py:meth:`.plot.manager.PlotManager.register_curve_tools`

            :py:meth:`.plot.manager.PlotManager.register_image_tools`

            :py:meth:`.plot.manager.PlotManager.register_all_image_tools`

            :py:meth:`.plot.manager.PlotManager.register_all_tools`
        """
        self.register_standard_tools()
        self.add_separator_tool()
        self.register_curve_tools()
        self.add_separator_tool()
        self.register_other_tools()
        self.add_separator_tool()
        self.update_tools_status()
        self.get_default_tool().activate()

    def register_all_image_tools(self) -> None:
        """
        Register standard, image-related and other tools

        .. seealso::

            :py:meth:`.plot.manager.PlotManager.add_tool`

            :py:meth:`.plot.manager.PlotManager.register_standard_tools`

            :py:meth:`.plot.manager.PlotManager.register_other_tools`

            :py:meth:`.plot.manager.PlotManager.register_curve_tools`

            :py:meth:`.plot.manager.PlotManager.register_image_tools`

            :py:meth:`.plot.manager.PlotManager.register_all_curve_tools`

            :py:meth:`.plot.manager.PlotManager.register_all_tools`
        """
        self.register_standard_tools()
        self.add_separator_tool()
        self.register_image_tools()
        self.add_separator_tool()
        self.register_other_tools()
        self.add_separator_tool()
        self.update_tools_status()
        self.get_default_tool().activate()

    def register_all_tools(self) -> None:
        """
        Register standard, curve and image-related and other tools

        .. seealso::

            :py:meth:`.plot.manager.PlotManager.add_tool`

            :py:meth:`.plot.manager.PlotManager.register_standard_tools`

            :py:meth:`.plot.manager.PlotManager.register_other_tools`

            :py:meth:`.plot.manager.PlotManager.register_curve_tools`

            :py:meth:`.plot.manager.PlotManager.register_image_tools`

            :py:meth:`.plot.manager.PlotManager.register_all_image_tools`

            :py:meth:`.plot.manager.PlotManager.register_all_curve_tools`
        """
        self.register_standard_tools()
        self.add_separator_tool()
        self.register_curve_tools()
        self.add_separator_tool()
        self.register_image_tools()
        self.add_separator_tool()
        self.register_other_tools()
        self.add_separator_tool()
        self.update_tools_status()
        self.get_default_tool().activate()

    def register_all_annotation_tools(self) -> None:
        """
        Register all annotation tools for the plot
        """
        self.add_separator_tool()
        self.add_tool(AnnotatedPointTool)
        self.add_tool(AnnotatedSegmentTool)
        self.add_tool(AnnotatedRectangleTool)
        self.add_tool(AnnotatedObliqueRectangleTool)
        self.add_tool(AnnotatedCircleTool)
        self.add_tool(AnnotatedEllipseTool)
        self.add_tool(LabelTool)

    def register_curve_annotation_tools(self) -> None:
        """
        Register all curve friendly annotation tools for the plot
        """
        self.add_separator_tool()
        self.add_tool(AnnotatedPointTool)
        self.add_tool(AnnotatedSegmentTool)
        self.add_tool(LabelTool)

    def register_image_annotation_tools(self) -> None:
        """
        Register all image friendly annotation tools for the plot
        """
        # No curve-specific annotation tool, so this is equivalent to the
        # register_all_annotation_tools function for now
        self.register_all_annotation_tools()

    def synchronize_axis(self, axis_id: int, plot_ids: list[str]) -> None:
        """
        Synchronize axis of plots

        Args:
            axis_id: axis id
            plot_ids: list of plot ids
        """
        for plot_id in plot_ids:
            synclist = self.synchronized_plots.setdefault(plot_id, [])
            for plot2_id in plot_ids:
                if plot_id == plot2_id:
                    continue
                item = (axis_id, plot2_id)
                if item not in synclist:
                    synclist.append(item)

    def plot_axis_changed(self, plot: BasePlot) -> None:
        """
        Plot axis changed, update other synchronized plots (if any)

        Args:
            plot: plot instance
        """
        plot_id = plot.plot_id
        if plot_id not in self.synchronized_plots:
            return
        for axis_id, other_plot_id in self.synchronized_plots[plot_id]:
            scalediv: QwtScaleDiv = plot.axisScaleDiv(axis_id)
            other_plot = self.get_plot(other_plot_id)
            lb = scalediv.lowerBound()
            ub = scalediv.upperBound()
            other_plot.setAxisScale(axis_id, lb, ub)
            other_plot.replot()


assert_interfaces_valid(PlotManager)



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""Visualisation tools for plotpy tests"""

from __future__ import annotations

from typing import TYPE_CHECKING

from plotpy.builder import make

if TYPE_CHECKING:
    from plotpy.items import BaseImageItem, CurveItem
    from plotpy.plot import PlotDialog


def show_items(
    items: list[CurveItem | BaseImageItem],
    plot_type: str = "auto",
    wintitle: str = "Plot items",
    title: str = "Title",
    xlabel: str = "X",
    ylabel: str = "Y",
    auto_tools: bool = True,
    lock_aspect_ratio: bool | None = None,
    curve_antialiasing: bool | None = None,
    show_itemlist: bool = True,
    show_contrast: bool = False,
    winsize: tuple[int, int] | None = None,
) -> PlotDialog:
    """Show plot items in a dialog box"""
    winsize = (640, 480) if winsize is None else winsize
    win = make.dialog(
        edit=False,
        toolbar=True,
        wintitle=wintitle,
        title=title,
        xlabel=xlabel,
        ylabel=ylabel,
        type=plot_type,
        auto_tools=auto_tools,
        lock_aspect_ratio=lock_aspect_ratio,
        curve_antialiasing=curve_antialiasing,
        show_itemlist=show_itemlist,
        show_contrast=show_contrast,
        size=winsize,
    )
    plot = win.manager.get_plot()
    for item in items:
        plot.add_item(item)
    plot.set_items_readonly(False)
    win.show()
    return win



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""
Plot coordinates
----------------

Overview
^^^^^^^^

The :mod:`plotpy.coords` module provides functions to convert coordinates
between canvas and axes coordinates systems.

The following functions are available:

* :py:func:`.canvas_to_axes`
* :py:func:`.axes_to_canvas`
* :py:func:`.pixelround`

Reference
^^^^^^^^^

.. autofunction:: canvas_to_axes
.. autofunction:: axes_to_canvas
.. autofunction:: pixelround
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from qtpy.QtCore import QPointF
    from qwt import QwtPlot, QwtPlotItem


def canvas_to_axes(item: QwtPlotItem, pos: QPointF) -> tuple[float, float] | None:
    """Convert position from canvas coordinates system to axes coordinates

    Args:
        item: Plot item
        pos: Position in canvas coordinates system

    Returns:
        Position in axes coordinates system or None if item is None
    """
    if item is None:
        return None
    plot: QwtPlot = item.plot()
    ax, ay = item.xAxis(), item.yAxis()
    return plot.invTransform(ax, pos.x()), plot.invTransform(ay, pos.y())


def axes_to_canvas(item: QwtPlotItem, x: float, y: float) -> tuple[float, float] | None:
    """Convert (x,y) from axes coordinates to canvas coordinates system

    Args:
        item: Plot item
        x: X position in axes coordinates system
        y: Y position in axes coordinates system

    Returns:
        Position in canvas coordinates system or None if item is None
    """
    if item is None:
        return None
    plot: QwtPlot = item.plot()
    ax, ay = item.xAxis(), item.yAxis()
    return plot.transform(ax, x), plot.transform(ay, y)


def pixelround(x: float, corner: str | None = None) -> int:
    """Get pixel index from pixel coordinate

    Args:
        x: Pixel coordinate
        corner: None (not a corner), 'TL' (top-left corner),
         'BR' (bottom-right corner)

    Returns:
        int: Pixel index
    """
    assert corner is None or corner in ("TL", "BR")
    if corner is None:
        return np.floor(x)
    elif corner == "BR":
        return np.ceil(x)
    elif corner == "TL":
        return np.floor(x)



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""Rotate/crop test: using the scaler C++ engine to rotate/crop images"""

# guitest: show

import numpy as np
from guidata.env import execenv
from guidata.qthelpers import exec_dialog, qt_app_context
from qtpy import QtWidgets as QW

from plotpy import io
from plotpy.builder import make
from plotpy.tests import get_path
from plotpy.widgets.rotatecrop import (
    MultipleRotateCropWidget,
    RotateCropDialog,
    RotateCropWidget,
)


def create_test_data(fname, func=None):
    """Create test data"""
    array0 = io.imread(get_path(fname), to_grayscale=True)
    if func is not None:
        array0 = func(array0)
    item0 = make.trimage(array0, dx=0.1, dy=0.1)
    return array0, item0


def widget_test(fname="brain.png"):
    """Test the rotate/crop widget"""
    _array0, item = create_test_data(fname)
    widget = RotateCropWidget(None, toolbar=True)
    widget.transform.set_item(item)
    widget.show()
    return widget


def multiple_widget_test(fname="brain.png"):
    """Test the multiple rotate/crop widget"""
    _array0, item0 = create_test_data(fname)
    _array1, item1 = create_test_data(fname, func=lambda arr: np.rot90(arr, 1))
    _array2, item2 = create_test_data(fname, func=lambda arr: np.rot90(arr, 2))
    widget = MultipleRotateCropWidget(None)
    widget.set_items(item0, item1, item2)
    widget.show()
    return widget


def imshow(data, title=None, hold=False):
    """Show image"""
    dlg = make.dialog(wintitle=title, type="image", size=(800, 600))
    dlg.manager.get_plot().add_item(make.image(data))
    if hold:
        dlg.show()
    else:
        exec_dialog(dlg)
    return dlg


def dialog_test(fname="brain.png"):
    """Test the rotate/crop dialog"""
    array0, item = create_test_data(fname)
    dlg = RotateCropDialog(None, toolbar=True)
    dlg.transform.set_item(item)
    if exec_dialog(dlg) == QW.QDialog.Accepted:
        array1 = dlg.transform.get_result()
        if array0.shape == array1.shape:
            assert (array1 == array0).all()
            imshow(array1 - array0, title="array1-array0")
        else:
            print(array0.shape, "-->", array1.shape)
        dlg1 = imshow(array0, title="array0")
        dlg2 = imshow(array1, title="array1")
        return dlg, dlg1, dlg2


def test_rotate_crop():
    """Test the flip/rotate widget and dialog"""
    persist_list = []
    with qt_app_context(exec_loop=not execenv.unattended):
        persist_list.append(multiple_widget_test("brain.png"))
    with qt_app_context(exec_loop=not execenv.unattended):
        persist_list.append(widget_test("brain.png"))
    with qt_app_context(exec_loop=False):
        persist_list.append(dialog_test("brain.png"))


if __name__ == "__main__":
    test_rotate_crop()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

# pylint: disable=C0103

"""
Annotation Item builder
-----------------------

This module provides a set of factory functions to simplify the creation
of annotation items.
"""

# Note: when adding method to builder classes, please do not forget to update the
# documentation (see builder.rst file). Because of class inheritance, the methods
# are not automatically documented (otherwise, they would be sorted alphabetically,
# due to a limitation of sphinx auto-doc).

from __future__ import annotations

import numpy as np

from plotpy.config import _
from plotpy.items import (
    AnnotatedCircle,
    AnnotatedEllipse,
    AnnotatedPoint,
    AnnotatedPolygon,
    AnnotatedRectangle,
    AnnotatedSegment,
)
from plotpy.styles import AnnotationParam


class AnnotationBuilder:
    """Class regrouping a set of factory functions to simplify the creation
    of annotation items."""

    def __get_annotationparam(
        self,
        title: str | None = None,
        subtitle: str | None = None,
        show_label: bool | None = None,
        show_computations: bool | None = None,
        show_subtitle: bool | None = None,
        format: str | None = None,
        uncertainty: float | None = None,
        transform_matrix: np.ndarray | None = None,
        readonly: bool | None = None,
        private: bool | None = None,
    ) -> AnnotationParam:
        param = AnnotationParam(_("Annotation"), icon="annotation.png")
        if title is not None:
            param.title = title
        if subtitle is not None:
            param.subtitle = subtitle
        if show_label is not None:
            param.show_label = show_label
        if show_computations is not None:
            param.show_computations = show_computations
        if show_subtitle is not None:
            param.show_subtitle = show_subtitle
        if format is not None:
            param.format = format
        if uncertainty is not None:
            param.uncertainty = uncertainty
        if transform_matrix is not None:
            param.transform_matrix = transform_matrix
        if readonly is not None:
            param.readonly = readonly
        if private is not None:
            param.private = private
        return param

    def annotated_point(
        self,
        x: float,
        y: float,
        title: str | None = None,
        subtitle: str | None = None,
        show_label: bool | None = None,
        show_computations: bool | None = None,
        show_subtitle: bool | None = None,
        format: str | None = None,
        uncertainty: float | None = None,
        transform_matrix: np.ndarray | None = None,
        readonly: bool | None = None,
        private: bool | None = None,
    ) -> AnnotatedPoint:
        """Make an annotated point `plot item`

        Args:
            x: point x coordinate
            y: point y coordinate
            title: label name. Default is None
            subtitle: label subtitle. Default is None
            show_label: show label. Default is None
            show_computations: show computations. Default is None
            show_subtitle: show subtitle. Default is None
            format: string formatting. Default is None
            uncertainty: measurement relative uncertainty. Default is None
            transform_matrix: transform matrix. Default is None
            readonly: readonly. Default is None
            private: private. Default is None

        Returns:
            :py:class:`.AnnotatedPoint` object
        """
        param = self.__get_annotationparam(
            title=title,
            subtitle=subtitle,
            show_label=show_label,
            show_computations=show_computations,
            show_subtitle=show_subtitle,
            format=format,
            uncertainty=uncertainty,
            transform_matrix=transform_matrix,
            readonly=readonly,
            private=private,
        )
        shape = AnnotatedPoint(x, y, param)
        shape.set_style("plot", "shape/drag")
        return shape

    def __annotated_shape(
        self,
        shapeclass,
        points,
        title,
        subtitle,
        show_label,
        show_computations,
        show_subtitle,
        format,
        uncertainty,
        transform_matrix,
        readonly,
        private,
    ):
        param = self.__get_annotationparam(
            title=title,
            subtitle=subtitle,
            show_label=show_label,
            show_computations=show_computations,
            show_subtitle=show_subtitle,
            format=format,
            uncertainty=uncertainty,
            transform_matrix=transform_matrix,
            readonly=readonly,
            private=private,
        )
        if isinstance(points, np.ndarray):
            shape = shapeclass(points, annotationparam=param)
        else:
            shape = shapeclass(*points, annotationparam=param)
        shape.set_style("plot", "shape/drag")
        return shape

    def annotated_rectangle(
        self,
        x0: float,
        y0: float,
        x1: float,
        y1: float,
        title: str | None = None,
        subtitle: str | None = None,
        show_label: bool | None = None,
        show_computations: bool | None = None,
        show_subtitle: bool | None = None,
        format: str | None = None,
        uncertainty: float | None = None,
        transform_matrix: np.ndarray | None = None,
        readonly: bool | None = None,
        private: bool | None = None,
    ) -> AnnotatedRectangle:
        """Make an annotated rectangle `plot item`

        Args:
            x0: rectangle x0 coordinate
            y0: rectangle y0 coordinate
            x1: rectangle x1 coordinate
            y1: rectangle y1 coordinate
            title: label name. Default is None
            subtitle: label subtitle. Default is None
            show_label: show label. Default is None
            show_computations: show computations. Default is None
            show_subtitle: show subtitle. Default is None
            format: string formatting. Default is None
            uncertainty: measurement relative uncertainty. Default is None
            transform_matrix: transform matrix. Default is None
            readonly: readonly. Default is None
            private: private. Default is None

        Returns:
            :py:class:`.AnnotatedRectangle` object
        """
        points = x0, y0, x1, y1
        return self.__annotated_shape(
            AnnotatedRectangle,
            points,
            title,
            subtitle,
            show_label,
            show_computations,
            show_subtitle,
            format,
            uncertainty,
            transform_matrix,
            readonly,
            private,
        )

    def annotated_ellipse(
        self,
        x0: float,
        y0: float,
        x1: float,
        y1: float,
        x2: float | None = None,
        y2: float | None = None,
        x3: float | None = None,
        y3: float | None = None,
        title: str | None = None,
        subtitle: str | None = None,
        show_label: bool | None = None,
        show_computations: bool | None = None,
        show_subtitle: bool | None = None,
        format: str | None = None,
        uncertainty: float | None = None,
        transform_matrix: np.ndarray | None = None,
        readonly: bool | None = None,
        private: bool | None = None,
    ) -> AnnotatedEllipse:
        """Make an annotated ellipse `plot item`

        Args:
            x0: ellipse x0 coordinate
            y0: ellipse y0 coordinate
            x1: ellipse x1 coordinate
            y1: ellipse y1 coordinate
            x2: ellipse x2 coordinate. Default is None
            y2: ellipse y2 coordinate. Default is None
            x3: ellipse x3 coordinate. Default is None
            y3: ellipse y3 coordinate. Default is None
            title: label name. Default is None
            subtitle: label subtitle. Default is None
            show_label: show label. Default is None
            show_computations: show computations. Default is None
            show_subtitle: show subtitle. Default is None
            format: string formatting. Default is None
            uncertainty: measurement relative uncertainty. Default is None
            transform_matrix: transform matrix. Default is None
            readonly: readonly. Default is None
            private: private. Default is None

        Returns:
            :py:class:`.AnnotatedEllipse` object
        """
        points = x0, y0, x1, y1
        item = self.__annotated_shape(
            AnnotatedEllipse,
            points,
            title,
            subtitle,
            show_label,
            show_computations,
            show_subtitle,
            format,
            uncertainty,
            transform_matrix,
            readonly,
            private,
        )
        if x2 is not None and y2 is not None and x3 is not None and y3 is not None:
            item.set_ydiameter(x2, y2, x3, y3)
        return item

    def annotated_circle(
        self,
        x0: float,
        y0: float,
        x1: float,
        y1: float,
        title: str | None = None,
        subtitle: str | None = None,
        show_label: bool | None = None,
        show_computations: bool | None = None,
        show_subtitle: bool | None = None,
        format: str | None = None,
        uncertainty: float | None = None,
        transform_matrix: np.ndarray | None = None,
        readonly: bool | None = None,
        private: bool | None = None,
    ) -> AnnotatedCircle:
        """Make an annotated circle `plot item`

        Args:
            x0: circle x0 coordinate
            y0: circle y0 coordinate
            x1: circle x1 coordinate
            y1: circle y1 coordinate
            title: label name. Default is None
            subtitle: label subtitle. Default is None
            show_label: show label. Default is None
            show_computations: show computations. Default is None
            show_subtitle: show subtitle. Default is None
            format: string formatting. Default is None
            uncertainty: measurement relative uncertainty. Default is None
            transform_matrix: transform matrix. Default is None
            readonly: readonly. Default is None
            private: private. Default is None

        Returns:
            :py:class:`.AnnotatedCircle` object
        """
        points = x0, y0, x1, y1
        return self.__annotated_shape(
            AnnotatedCircle,
            points,
            title,
            subtitle,
            show_label,
            show_computations,
            show_subtitle,
            format,
            uncertainty,
            transform_matrix,
            readonly,
            private,
        )

    def annotated_segment(
        self,
        x0: float,
        y0: float,
        x1: float,
        y1: float,
        title: str | None = None,
        subtitle: str | None = None,
        show_label: bool | None = None,
        show_computations: bool | None = None,
        show_subtitle: bool | None = None,
        format: str | None = None,
        uncertainty: float | None = None,
        transform_matrix: np.ndarray | None = None,
        readonly: bool | None = None,
        private: bool | None = None,
    ) -> AnnotatedSegment:
        """Make an annotated segment `plot item`

        Args:
            x0: segment x0 coordinate
            y0: segment y0 coordinate
            x1: segment x1 coordinate
            y1: segment y1 coordinate
            title: label name. Default is None
            subtitle: label subtitle. Default is None
            show_label: show label. Default is None
            show_computations: show computations. Default is None
            show_subtitle: show subtitle. Default is None
            format: string formatting. Default is None
            uncertainty: measurement relative uncertainty. Default is None
            transform_matrix: transform matrix. Default is None
            readonly: readonly. Default is None
            private: private. Default is None

        Returns:
            :py:class:`.AnnotatedSegment` object
        """
        points = x0, y0, x1, y1
        return self.__annotated_shape(
            AnnotatedSegment,
            points,
            title,
            subtitle,
            show_label,
            show_computations,
            show_subtitle,
            format,
            uncertainty,
            transform_matrix,
            readonly,
            private,
        )

    def annotated_polygon(
        self,
        points: np.ndarray,
        title: str | None = None,
        subtitle: str | None = None,
        show_label: bool | None = None,
        show_computations: bool | None = None,
        show_subtitle: bool | None = None,
        format: str | None = None,
        uncertainty: float | None = None,
        transform_matrix: np.ndarray | None = None,
        readonly: bool | None = None,
        private: bool | None = None,
    ) -> AnnotatedPolygon:
        """Make an annotated polygon `plot item`

        Args:
            points: polygon points
            title: label name. Default is None
            subtitle: label subtitle. Default is None
            show_label: show label. Default is None
            show_computations: show computations. Default is None
            show_subtitle: show subtitle. Default is None
            format: string formatting. Default is None
            uncertainty: measurement relative uncertainty. Default is None
            transform_matrix: transform matrix. Default is None
            readonly: readonly. Default is None
            private: private. Default is None

        Returns:
            :py:class:`.AnnotatedPolygon` object
        """
        return self.__annotated_shape(
            AnnotatedPolygon,
            points,
            title,
            subtitle,
            show_label,
            show_computations,
            show_subtitle,
            format,
            uncertainty,
            transform_matrix,
            readonly,
            private,
        )



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""PolygonMapItem test

PolygonMapItem is intended to display maps ie items containing
several hundreds of independent polygons.
"""

# guitest: show

import numpy as np
from guidata.qthelpers import qt_app_context

from plotpy.items import PolygonMapItem
from plotpy.plot import PlotDialog, PlotOptions

# Create a sample dataset consisting of tesselated circles randomly placed
# in a box
RMAX = 0.5
XMAX = YMAX = 10.0
NSEGMIN = 4
NSEGMAX = 300


def create_circle():
    x, y, rmax = np.random.rand(3)
    rmax *= RMAX
    x *= XMAX
    y *= YMAX
    nseg = np.random.randint(NSEGMIN, NSEGMAX)
    th = np.linspace(0, 2 * np.pi, nseg)
    PTS = np.empty((nseg, 2), float)
    PTS[:, 0] = x + rmax * np.cos(th)
    PTS[:, 1] = y + rmax * np.sin(th)
    return PTS


NCIRC = 1000
COLORS = [
    (0xFF000000, 0x8000FF00),
    (0xFF0000FF, 0x800000FF),
    (0xFF000000, 0x80FF0000),
    (0xFF00FF00, 0x80000000),
]


def test_polygons():
    """Test"""
    with qt_app_context(exec_loop=True):
        win = PlotDialog(
            edit=True,
            toolbar=True,
            title="Sample multi-polygon item",
            options=PlotOptions(show_itemlist=True),
        )
        plot = win.manager.get_plot()
        plot.set_aspect_ratio(lock=True)
        plot.set_antialiasing(False)
        plot.set_axis_direction("left", False)
        plot.set_axis_title("bottom", "Lon")
        plot.set_axis_title("left", "Lat")

        points = []
        offsets = np.zeros((NCIRC, 2), np.int32)
        colors = np.zeros((NCIRC, 2), np.uint32)
        npts = 0
        for k in range(NCIRC):
            pts = create_circle()
            offsets[k, 0] = k
            offsets[k, 1] = npts
            npts += pts.shape[0]
            points.append(pts)
            colors[k, 0] = COLORS[k % len(COLORS)][0]
            colors[k, 1] = COLORS[(3 * k) % len(COLORS)][1]
        points = np.concatenate(points)

        print(NCIRC, "Polygons")
        print(points.shape[0], "Points")

        crv = PolygonMapItem()
        crv.set_data(points, offsets, colors)
        plot.add_item(crv, z=0)
        win.show()


if __name__ == "__main__":
    test_polygons()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see guidata/LICENSE for details)

"""
Test dark/light theme switching
"""

from __future__ import annotations

import sys
from typing import Literal

import pytest
from guidata import qthelpers as qth

try:
    # guidata > V2.6.2
    from guidata.tests.widgets.test_theme import ColorModeWidget as GuidataWidget
except ImportError:
    # guidata V2.6.2
    from guidata.tests.widgets.test_theme import (
        TestWidget as GuidataWidget,  # noqa: F401
    )
from qtpy import QtWidgets as QW

from plotpy.builder import make
from plotpy.config import update_plotpy_color_mode
from plotpy.plot import PlotOptions, PlotWidget
from plotpy.tests import data as ptd


class ColorModeWidget(GuidataWidget):
    """Testing color mode switching for PlotPy and guidata widgets"""

    SIZE = (1400, 600)

    def __init__(self, default: Literal["light", "dark", "auto"] = qth.AUTO) -> None:
        self.plot_widget: PlotWidget | None = None
        super().__init__(default=default)

    def setup_widgets(self):
        """Setup widgets"""
        super().setup_widgets()
        options = PlotOptions(type="image", show_contrast=True)
        self.plot_widget = widget = PlotWidget(self, options=options)
        plot = self.plot_widget.get_plot()
        item = make.image(ptd.gen_image4(300, 200))
        plot.add_item(item)
        plot.set_active_item(item, select=False)
        widget.setSizePolicy(QW.QSizePolicy.Expanding, QW.QSizePolicy.Expanding)
        self.grid_layout.addWidget(widget, 1, 2)

    def change_color_mode(self, mode: str) -> None:
        """Change color mode"""
        super().change_color_mode(mode)
        update_plotpy_color_mode()


@pytest.mark.skipif(reason="Not suitable for automated testing")
def test_dark_light_themes(
    default: Literal["light", "dark", "auto"] | None = None,
) -> None:
    """Test dark/light theme switching"""
    with qth.qt_app_context(exec_loop=True):
        widget = ColorModeWidget(default=qth.AUTO if default is None else default)
        widget.show()


if __name__ == "__main__":
    test_dark_light_themes(None if len(sys.argv) < 2 else sys.argv[1])



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""Reverse y-axis test for curve plotting"""

# guitest: show

import numpy as np
from guidata.qthelpers import qt_app_context

from plotpy.builder import make
from plotpy.tests import vistools as ptv


def test_plot_yreverse():
    """Testing reverse x/y axes"""
    with qt_app_context(exec_loop=True):
        x = np.linspace(-10, 10, 200)
        win = ptv.show_items(
            [make.curve(x, x * np.exp(-x), color="b")],
            plot_type="curve",
            wintitle=test_plot_yreverse.__doc__,
        )
        plot = win.manager.get_plot()
        plot.set_axis_direction("left", True)
        plot.set_axis_direction("bottom", True)


if __name__ == "__main__":
    test_plot_yreverse()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""Test PlotBuilder shape factory methods"""

import numpy as np
import pytest
from guidata.qthelpers import qt_app_context

from plotpy.builder import make
from plotpy.tests import get_path
from plotpy.tests.unit.test_builder_curve import show_items_qtbot

DEFAULT_ARGS = {
    make.segment: [0.0, 0.0, 1.0, 1.0],
    make.rectangle: [0.0, 0.0, 1.0, 1.0],
    make.circle: [0.0, 0.0, 1.0, 1.0],
    make.ellipse: [0.0, 0.0, 1.0, 1.0, 0.0, 0.0, 1.0, 1.0],
}


def _make_standard_shape(
    method,
    title: str = None,
):
    """Make annotation"""
    args = DEFAULT_ARGS[method]
    return method(
        *args,
        title=title,
    )


@pytest.mark.parametrize(
    "method",
    [make.segment, make.rectangle, make.circle, make.ellipse],
)
def test_builder_standard_shape(method):
    with qt_app_context(exec_loop=False):
        items = []
        items.append(_make_standard_shape(method, title="title"))
        show_items_qtbot(items)


def test_builder_polygon():
    items = []
    x = np.linspace(0, 1, 10)
    y = x**2
    with qt_app_context(exec_loop=False):
        for closed in [True, False]:
            items.append(make.polygon(x, y, closed=closed, title="title"))
        show_items_qtbot(items)


def test_builder_svgshape():
    items = []
    svg_path = get_path("svg_target.svg")
    with open(svg_path, "rb") as f:
        svg_data = f.read()
    with qt_app_context(exec_loop=False):
        for shape_str in ("circle", "rectangle", "square"):
            for data_or_path in (svg_data, svg_path):
                items.append(
                    make.svg(shape_str, data_or_path, 0.0, 0.0, 1.0, 1.0, title="title")
                )
        show_items_qtbot(items)



# -*- coding: utf-8 -*-
from guidata.dataset import DataSet, TextItem
from qtpy import QtCore as QC

from plotpy.config import _
from plotpy.events import ClickHandler, setup_standard_tool_filter
from plotpy.tools.base import DefaultToolbarID, InteractiveTool


class LabelTool(InteractiveTool):
    """ """

    TITLE = _("Label")
    ICON = "label.png"
    LABEL_STYLE_SECT = "plot"
    LABEL_STYLE_KEY = "label"
    SWITCH_TO_DEFAULT_TOOL = True

    def __init__(
        self,
        manager,
        handle_label_cb=None,
        label_style=None,
        toolbar_id=DefaultToolbarID,
        title=None,
        icon=None,
        tip=None,
        switch_to_default_tool=None,
    ):
        self.handle_label_cb = handle_label_cb
        super().__init__(
            manager,
            toolbar_id,
            title=title,
            icon=icon,
            tip=tip,
            switch_to_default_tool=switch_to_default_tool,
        )
        if label_style is not None:
            self.label_style_sect = label_style[0]
            self.label_style_key = label_style[1]
        else:
            self.label_style_sect = self.LABEL_STYLE_SECT
            self.label_style_key = self.LABEL_STYLE_KEY

    def set_label_style(self, label):
        """

        :param label:
        """
        label.set_style(self.label_style_sect, self.label_style_key)

    def setup_filter(self, baseplot):
        """

        :param baseplot:
        :return:
        """
        filter = baseplot.filter
        start_state = filter.new_state()
        handler = ClickHandler(
            filter, QC.Qt.MouseButton.LeftButton, start_state=start_state
        )
        handler.SIG_CLICK_EVENT.connect(self.add_label_to_plot)
        return setup_standard_tool_filter(filter, start_state)

    def add_label_to_plot(self, filter, event):
        """

        :param filter:
        :param event:
        """
        plot = filter.plot

        # The following import is here to avoid circular imports
        # pylint: disable=import-outside-toplevel
        from plotpy.builder import make

        class TextParam(DataSet):
            text = TextItem("", _("Label"))

        textparam = TextParam(_("Label text"), icon=self.ICON)
        if textparam.edit(plot):
            text = textparam.text.replace("\n", "<br>")

            label = make.label(text, (0, 0), (10, 10), "TL")
            title = label.labelparam.label
            self.set_label_style(label)
            label.setTitle(self.TITLE)
            x = plot.invTransform(label.xAxis(), event.pos().x())
            y = plot.invTransform(label.yAxis(), event.pos().y())
            label.set_pos(x, y)
            label.setTitle(title)
            label.labelparam.update_param(label)
            plot.add_item(label)
            if self.handle_label_cb is not None:
                self.handle_label_cb(label)
            plot.replot()
            self.SIG_TOOL_JOB_FINISHED.emit()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""Styles unit tests"""

import gettext
import unittest

from qtpy import QtCore as QC
from qtpy import QtGui as QG
from qwt import QwtSymbol

from plotpy.config import UserConfig, _
from plotpy.styles.base import LineStyleParam, SymbolParam

gettext.install("test")
CONF = UserConfig({})
CONF.set_application("plotpy", version="0.0.0", load=False)


class TestSymbol(unittest.TestCase):
    def test_default(self):
        sym = SymbolParam(_("Symbol"))
        _obj = sym.build_symbol()

    def test_update(self):
        obj = QwtSymbol(
            QwtSymbol.Rect,
            QG.QBrush(QC.Qt.black),
            QG.QPen(QC.Qt.yellow),
            QC.QSize(3, 3),
        )
        sym = SymbolParam(_("Symbol"))
        sym.update_param(obj)
        self.assertEqual(sym.marker, "Rect")
        self.assertEqual(sym.size, 3)
        self.assertEqual(sym.edgecolor, "#ffff00")
        self.assertEqual(sym.facecolor, "#000000")

    def test_saveconfig(self):
        sym = SymbolParam(_("Symbol"))
        sym.write_config(CONF, "sym", "")
        sym = SymbolParam(_("Symbol"))
        sym.read_config(CONF, "sym", "")

    def test_changeconfig(self):
        obj = QwtSymbol(
            QwtSymbol.Rect,
            QG.QBrush(QC.Qt.black),
            QG.QPen(QC.Qt.yellow),
            QC.QSize(3, 3),
        )
        sym = SymbolParam(_("Symbol"))
        sym.update_param(obj)
        sym.write_config(CONF, "sym", "")
        sym = SymbolParam(_("Symbol"))
        sym.read_config(CONF, "sym", "")
        self.assertEqual(sym.marker, "Rect")
        self.assertEqual(sym.size, 3)
        self.assertEqual(sym.edgecolor, "#ffff00")
        self.assertEqual(sym.facecolor, "#000000")
        sym.build_symbol()


class TestLineStyle(unittest.TestCase):
    def test_default(self):
        ls = LineStyleParam(_("Line style"))
        _obj = ls.build_pen()

    def test_update(self):
        obj = QG.QPen(QC.Qt.red, 2, QC.Qt.SolidLine)
        ls = LineStyleParam(_("Line style"))
        ls.update_param(obj)
        self.assertEqual(ls.width, 2)
        self.assertEqual(ls.style, "SolidLine")
        self.assertEqual(ls.color, "#ff0000")

    def test_saveconfig(self):
        ls = LineStyleParam(_("Line style"))
        ls.write_config(CONF, "ls", "")
        ls = LineStyleParam(_("Line style"))
        ls.read_config(CONF, "ls", "")

    def test_changeconfig(self):
        obj = QG.QPen(QC.Qt.red, 2, QC.Qt.SolidLine)
        ls = LineStyleParam(_("Line style"))
        ls.update_param(obj)
        ls.write_config(CONF, "ls", "")
        ls = LineStyleParam(_("Line style"))
        ls.read_config(CONF, "ls", "")
        self.assertEqual(ls.width, 2)
        self.assertEqual(ls.style, "SolidLine")
        self.assertEqual(ls.color, "#ff0000")


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""Shows how to customize a shape created with a tool like RectangleTool"""

# guitest: show

from guidata.qthelpers import qt_app_context

from plotpy.builder import make
from plotpy.constants import LUTAlpha
from plotpy.styles import style_generator, update_style_attr
from plotpy.tests import get_path
from plotpy.tools import (
    EllipseTool,
    MultiLineTool,
    PolygonTool,
    RectangleTool,
    SegmentTool,
)

STYLE = style_generator()


def customize_shape(shape):
    global STYLE
    param = shape.shapeparam
    style = next(STYLE)
    update_style_attr(style, param)
    param.update_item(shape)
    shape.plot().replot()


def create_window():
    gridparam = make.gridparam(
        background="black", minor_enabled=(False, False), major_style=(".", "gray", 1)
    )
    win = make.dialog(
        edit=False,
        toolbar=True,
        wintitle="All image and plot tools test",
        gridparam=gridparam,
        type="image",
        size=(800, 600),
    )
    for toolklass in (
        RectangleTool,
        EllipseTool,
        SegmentTool,
        MultiLineTool,
        PolygonTool,
    ):
        win.manager.add_tool(toolklass, handle_final_shape_cb=customize_shape)
    return win


def test_customize_shape_tool():
    """Test"""
    with qt_app_context(exec_loop=True):
        filename = get_path("brain.png")
        win = create_window()
        image = make.image(
            filename=filename, colormap="bone", alpha_function=LUTAlpha.LINEAR
        )
        plot = win.manager.get_plot()
        plot.add_item(image)
        win.show()


if __name__ == "__main__":
    test_customize_shape_tool()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

# pylint: disable=C0103

"""
Interactive plot
----------------

The `pyplot` module provides an interactive plotting interface similar to
`Matplotlib`'s, i.e. with MATLAB-like syntax.
"""

import sys

import guidata
import numpy as np
from guidata.configtools import get_icon
from guidata.env import execenv
from guidata.qthelpers import win32_fix_title_bar_background
from qtpy import QtCore as QC
from qtpy import QtGui as QG
from qtpy import QtWidgets as QW
from qtpy.QtPrintSupport import QPrinter

from plotpy import io
from plotpy._scaler import INTERP_AA, INTERP_LINEAR, INTERP_NEAREST
from plotpy.builder import make
from plotpy.config import _
from plotpy.mathutils.colormap import ALL_COLORMAPS
from plotpy.panels.contrastadjustment import ContrastAdjustment
from plotpy.panels.csection.cswidget import XCrossSection, YCrossSection
from plotpy.panels.itemlist import PlotItemList
from plotpy.plot.base import BasePlot, BasePlotOptions
from plotpy.plot.manager import PlotManager

_qapp = None
_interactive = False
_figures = {}
_current_fig = None
_current_axes = None


def create_qapplication(exec_loop=False):
    """Creating Qt application (only once) and eventually exec Qt main loop"""
    global _qapp  # pylint: disable=global-statement
    if _qapp is None:
        _qapp = guidata.qapplication()
    if exec_loop and not execenv.unattended:
        _qapp.exec()


class Window(QW.QMainWindow):
    """Figure window"""

    def __init__(self, title):
        super().__init__()
        win32_fix_title_bar_background(self)

        self.default_tool = None
        self.plots = []
        self.itemlist = PlotItemList(None)
        self.contrast = ContrastAdjustment(None)
        self.xcsw = XCrossSection(None)
        self.ycsw = YCrossSection(None)

        self.manager = PlotManager(self)
        self.toolbar = QW.QToolBar(_("Tools"), self)
        self.manager.add_toolbar(self.toolbar, "default")
        self.toolbar.setMovable(True)
        self.toolbar.setFloatable(True)
        self.addToolBar(QC.Qt.TopToolBarArea, self.toolbar)

        frame = QW.QFrame(self)
        self.setCentralWidget(frame)
        self.layout = QW.QGridLayout()
        layout = QW.QVBoxLayout(frame)
        frame.setLayout(layout)
        layout.addLayout(self.layout)
        self.frame = frame

        self.setWindowTitle(title)
        self.setWindowIcon(get_icon("plotpy.svg"))

    def closeEvent(self, event):
        """Reimplementing QWidget.closeEvent

        :param event:
        """
        global _figures, _current_fig, _current_axes
        figure_title = str(self.windowTitle())
        if _figures.pop(figure_title) == _current_fig:
            _current_fig = None
            _current_axes = None
        self.itemlist.close()
        self.contrast.close()
        self.xcsw.close()
        self.ycsw.close()
        event.accept()

    def add_plot(self, i, j, plot):
        """Add plot to window

        :param i:
        :param j:
        :param plot:
        """
        i_int = int(i)
        j_int = int(j)
        self.layout.addWidget(plot, i_int, j_int)
        self.manager.add_plot(plot)
        self.plots.append(plot)

    def replot(self):
        """Replot all plots"""
        for plot in self.plots:
            plot.replot()
            item = plot.get_default_item()
            if item is not None:
                plot.set_active_item(item)
                item.unselect()

    def add_panels(self, images=False):
        """Add panels

        :param images:
        """
        self.manager.add_panel(self.itemlist)
        if images:
            for panel in (self.ycsw, self.xcsw, self.contrast):
                panel.hide()
                self.manager.add_panel(panel)

    def register_tools(self, images=False):
        """Register tools

        :param images:
        """
        if images:
            self.manager.register_all_image_tools()
        else:
            self.manager.register_all_curve_tools()

    def display(self):
        """Display window"""
        self.show()
        self.replot()
        self.manager.get_default_tool().activate()
        self.manager.update_tools_status()


class Figure:
    """Object representing a figure"""

    def __init__(self, title):
        self.axes = {}
        self.title = title
        self.win = None

    def get_axes(self, i, j):
        """Get axes

        :param i:
        :param j:
        :return:
        """
        if (i, j) in self.axes:
            return self.axes[(i, j)]

        ax = Axes()
        self.axes[(i, j)] = ax
        return ax

    def build_window(self):
        """Build window"""
        create_qapplication()
        self.win = Window(title=self.title)
        images = False
        for (i, j), ax in list(self.axes.items()):
            ax.setup_window(i, j, self.win)
            if ax.images:
                images = True
        self.win.add_panels(images=images)
        self.win.register_tools(images=images)

    def show(self):
        """Show window"""
        if not self.win:
            self.build_window()
        self.win.display()

    def save(self, fname, format, draft):
        """Save figure

        :param fname:
        :param format:
        :param draft:
        """
        if isinstance(fname, str):
            if format == "pdf":
                try:
                    mode = QPrinter.HighResolution
                except AttributeError:
                    # Some PySide6 / PyQt6 versions do not have this attribute on Linux
                    mode = QPrinter.ScreenResolution
                if draft:
                    mode = QPrinter.ScreenResolution
                printer = QPrinter(mode)
                try:
                    printer.setOutputFormat(QPrinter.PdfFormat)
                except AttributeError:
                    # PyQt6 on Linux
                    printer.setPrinterName("")
                printer.setPageOrientation(QG.QPageLayout.Landscape)
                printer.setOutputFileName(fname)
                printer.setCreator("plotpy.pyplot")
                self.print_(printer)
            else:
                if self.win is None:
                    self.show()
                pixmap = self.win.centralWidget().grab()
                pixmap.save(fname, format.upper())
        else:
            # Buffer
            fd = fname
            assert hasattr(fd, "write"), "object is not file-like as expected"
            if self.win is None:
                self.show()
            pixmap = self.win.centralWidget().grab()
            buff = QC.QBuffer()
            buff.open(QC.QIODevice.ReadWrite)
            pixmap.save(buff, format.upper())
            fd.write(buff.data())
            buff.close()
            fd.seek(0)

    def print_(self, device):
        """Print figure

        :param device:
        """
        if not self.win:
            self.build_window()
        W = device.width()
        H = device.height()

        coords = np.array(list(self.axes.keys()))
        imin = coords[:, 0].min()
        imax = coords[:, 0].max()
        jmin = coords[:, 1].min()
        jmax = coords[:, 1].max()
        w = W / (jmax - jmin + 1)
        h = H / (imax - imin + 1)
        paint = QG.QPainter(device)
        for (i, j), ax in list(self.axes.items()):
            oy = (i - imin) * h
            ox = (j - jmin) * w
            ax.widget.print_(paint, QC.QRect(ox, oy, w, h))


class Axes:
    """Object handling plots, images and axes properties"""

    def __init__(self):
        self.plots = []
        self.images = []
        self.last = None
        self.legend_position = None
        self.grid = False
        self.xlabel = ("", "")
        self.ylabel = ("", "")
        self.xcolor = ("black", "black")  # axis label colors
        self.ycolor = ("black", "black")  # axis label colors
        self.zlabel = None
        self.yreverse = False
        self.colormap = "jet"
        self.xscale = "lin"
        self.yscale = "lin"
        self.xlimits = None
        self.ylimits = None
        self.widget = None
        self.main_widget = None

    def add_legend(self, position):
        """Add legend

        :param position:
        """
        self.legend_position = position

    def set_grid(self, grid):
        """Set grid

        :param grid:
        """
        self.grid = grid

    def set_xlim(self, xmin, xmax):
        """Set x limits

        :param xmin:
        :param xmax:
        """
        self.xlimits = xmin, xmax
        self._update_plotwidget()

    def set_ylim(self, ymin, ymax):
        """Set y limits

        :param ymin:
        :param ymax:
        """
        self.ylimits = ymin, ymax
        self._update_plotwidget()

    def add_plot(self, item):
        """Add plot

        :param item:
        """
        self.plots.append(item)
        self.last = item

    def add_image(self, item):
        """Add image

        :param item:
        """
        self.images.append(item)
        self.last = item

    def setup_window(self, i, j, win):
        """Setup window

        :param i:
        :param j:
        :param win:
        """
        if self.images:
            plot = self.setup_image(i, j, win)
        else:
            plot = self.setup_plot(i, j, win)
        self.widget = plot
        plot.do_autoscale()
        self._update_plotwidget()

    def _update_plotwidget(self):
        """Update plot widget"""
        p = self.main_widget
        if p is None:
            return
        if self.grid:
            p.gridparam.maj_xenabled = True
            p.gridparam.maj_yenabled = True
            p.gridparam.update_grid(p)
        p.set_axis_color("bottom", self.xcolor[0])
        p.set_axis_color("top", self.xcolor[1])
        p.set_axis_color("left", self.ycolor[0])
        p.set_axis_color("right", self.ycolor[1])
        if self.xlimits is not None:
            p.set_axis_limits("bottom", *self.xlimits)
        if self.ylimits is not None:
            p.set_axis_limits("left", *self.ylimits)

    def setup_image(self, i, j, win: Window):
        """Setup image

        :param i:
        :param j:
        :param win:
        :return:
        """
        options = BasePlotOptions(
            xlabel=self.xlabel,
            ylabel=self.ylabel,
            yreverse=self.yreverse,
            type="image",
        )
        p = BasePlot(win, options=options)
        self.main_widget = p
        win.add_plot(i, j, p)
        for item in self.images + self.plots:
            if item in self.images:
                item.set_color_map(self.colormap)
            p.add_item(item)
        if self.legend_position is not None:
            p.add_item(make.legend(self.legend_position))
        return p

    def setup_plot(self, i, j, win):
        """Setup plot

        :param i:
        :param j:
        :param win:
        :return:
        """
        options = BasePlotOptions(xlabel=self.xlabel, ylabel=self.ylabel, type="curve")
        p = BasePlot(win, options=options)
        self.main_widget = p
        win.add_plot(i, j, p)
        for item in self.plots:
            p.add_item(item)
        p.enable_used_axes()
        active_item = p.get_active_item(force=True)
        p.set_scales(self.xscale, self.yscale)
        active_item.unselect()
        if self.legend_position is not None:
            p.add_item(make.legend(self.legend_position))
        return p


def _make_figure_title(N=None):
    """Make figure title"""
    global _figures
    if N is None:
        N = len(_figures) + 1
    if isinstance(N, str):
        return N
    else:
        return f"Figure {N:d}"


def figure(N=None):
    """Create a new figure"""
    global _figures, _current_fig, _current_axes
    title = _make_figure_title(N)
    if title in _figures:
        f = _figures[title]
    else:
        f = Figure(title)
        _figures[title] = f
    _current_fig = f
    _current_axes = None
    return f


def gcf():
    """Get current figure"""
    global _current_fig
    if _current_fig:
        return _current_fig
    else:
        return figure()


def gca():
    """Get current axes"""
    create_qapplication()  # Necessary because of setIcon call during item creation
    global _current_axes
    if not _current_axes:
        axes = gcf().get_axes(1, 1)
        _current_axes = axes
    return _current_axes


def show(mainloop=True):
    """
    Show all figures and enter Qt event loop
    This should be the last line of your script
    """
    global _figures, _interactive, _current_fig
    for fig in list(_figures.values()):
        fig.show()
    if not _interactive:
        if not _current_fig:
            print("Warning: must create a figure before showing it", file=sys.stderr)
        elif mainloop:
            create_qapplication(exec_loop=True)


def _show_if_interactive():
    """Show all figures if interactive mode is on"""
    global _interactive
    if _interactive:
        show()


def subplot(n, m, k):
    """
    Create a subplot command

    Example::

        import numpy as np
        x = np.linspace(-5, 5, 1000)
        figure(1)
        subplot(2, 1, 1)
        plot(x, np.sin(x), "r+")
        subplot(2, 1, 2)
        plot(x, np.cos(x), "g-")
        show()
    """
    global _current_axes
    lig = (k - 1) // m
    col = (k - 1) % m
    fig = gcf()
    axe = fig.get_axes(lig, col)
    _current_axes = axe
    return axe


def plot(*args, **kwargs):
    """
    Plot curves

    Example::

        import numpy as np
        x = np.linspace(-5, 5, 1000)
        plot(x, np.sin(x), "r+")
        plot(x, np.cos(x), "g-")
        show()
    """
    axe = gca()
    curves = make.mcurve(*args, **kwargs)
    if not isinstance(curves, list):
        curves = [curves]
    for curve in curves:
        axe.add_plot(curve)
    _show_if_interactive()
    return curves


def plotyy(x1, y1, x2, y2):
    """
    Plot curves with two different y axes

    Example::

        import numpy as np
        x = np.linspace(-5, 5, 1000)
        plotyy(x, np.sin(x), x, np.cos(x))
        ylabel("sinus", "cosinus")
        show()
    """
    axe = gca()
    curve1 = make.mcurve(x1, y1, yaxis="left")
    curve2 = make.mcurve(x2, y2, yaxis="right")
    axe.ycolor = (curve1.param.line.color, curve2.param.line.color)
    axe.add_plot(curve1)
    axe.add_plot(curve2)
    _show_if_interactive()
    return [curve1, curve2]


def hist(data, bins=None, logscale=None, title=None, color=None):
    """
    Plot 1-D histogram

    Example::

        from numpy.random import normal
        data = normal(0, 1, (2000, ))
        hist(data)
        show()
    """
    axe = gca()
    curve = make.histogram(
        data, bins=bins, logscale=logscale, title=title, color=color, yaxis="left"
    )
    axe.add_plot(curve)
    _show_if_interactive()
    return [curve]


def semilogx(*args, **kwargs):
    """
    Plot curves with logarithmic x-axis scale

    Example::

        import numpy as np
        x = np.linspace(-5, 5, 1000)
        semilogx(x, np.sin(12*x), "g-")
        show()
    """
    axe = gca()
    axe.xscale = "log"
    curve = make.mcurve(*args, **kwargs)
    axe.add_plot(curve)
    _show_if_interactive()
    return [curve]


def semilogy(*args, **kwargs):
    """
    Plot curves with logarithmic y-axis scale

    Example::

        import numpy as np
        x = np.linspace(-5, 5, 1000)
        semilogy(x, np.sin(12*x), "g-")
        show()
    """
    axe = gca()
    axe.yscale = "log"
    curve = make.mcurve(*args, **kwargs)
    axe.add_plot(curve)
    _show_if_interactive()
    return [curve]


def loglog(*args, **kwargs):
    """
    Plot curves with logarithmic x-axis and y-axis scales

    Example::

        import numpy as np
        x = np.linspace(-5, 5, 1000)
        loglog(x, np.sin(12*x), "g-")
        show()
    """
    axe = gca()
    axe.xscale = "log"
    axe.yscale = "log"
    curve = make.mcurve(*args, **kwargs)
    axe.add_plot(curve)
    _show_if_interactive()
    return [curve]


def errorbar(*args, **kwargs):
    """
    Plot curves with error bars

    Example::

        import numpy as np
        x = np.linspace(-5, 5, 1000)
        errorbar(x, -1+x**2/20+.2*np.random.rand(len(x)), x/20)
        show()
    """
    axe = gca()
    curve = make.merror(*args, **kwargs)
    axe.add_plot(curve)
    _show_if_interactive()
    return [curve]


def imread(fname, to_grayscale=False):
    """Read data from *fname*"""
    return io.imread(fname, to_grayscale=to_grayscale)


def imshow(data, interpolation=None, mask=None):
    """
    Display the image in *data* to current axes
    interpolation: 'nearest', 'linear' (default), 'antialiasing'

    Example::

        import numpy as np
        x = np.linspace(-5, 5, 1000)
        img = np.fromfunction(lambda x, y: np.sin((x/200.)*(y/200.)**2), (1000, 1000))
        gray()
        imshow(img)
        show()
    """
    axe = gca()

    if isinstance(data, np.ma.MaskedArray) and mask is None:
        mask = data.mask
        data = data.data
    if mask is None:
        img = make.image(data)
    else:
        img = make.maskedimage(data, mask, show_mask=True)
    if interpolation is not None:
        interp_dict = {
            "nearest": INTERP_NEAREST,
            "linear": INTERP_LINEAR,
            "antialiasing": INTERP_AA,
        }
        assert interpolation in interp_dict, "invalid interpolation option"
        img.set_interpolation(interp_dict[interpolation], size=5)
    axe.add_image(img)
    axe.yreverse = True
    _show_if_interactive()
    return [img]


def pcolor(*args):
    """
    Create a pseudocolor plot of a 2-D array

    Example::

        import numpy as np
        r = np.linspace(1., 16, 100)
        th = np.linspace(0., np.pi, 100)
        R, TH = np.meshgrid(r, th)
        X = R*np.cos(TH)
        Y = R*np.sin(TH)
        Z = 4*TH+R
        pcolor(X, Y, Z)
        show()
    """
    axe = gca()
    img = make.pcolor(*args)
    axe.add_image(img)
    axe.yreverse = len(args) == 1
    _show_if_interactive()
    return [img]


def interactive(state):
    """Toggle interactive mode"""
    global _interactive
    _interactive = state


def ion():
    """Turn interactive mode on"""
    interactive(True)


def ioff():
    """Turn interactive mode off"""
    interactive(False)


# TODO: The following functions (title, xlabel, ...) should update an already
#      shown figure to be compatible with interactive mode -- for now it just
#      works if these functions are called before showing the figure
def title(text):
    """Set current figure title"""
    global _figures
    fig = gcf()
    _figures.pop(fig.title)
    fig.title = text
    _figures[text] = fig


def xlabel(bottom="", top=""):
    """Set current x-axis label"""
    assert isinstance(bottom, str) and isinstance(top, str)
    axe = gca()
    axe.xlabel = (bottom, top)


def ylabel(left="", right=""):
    """Set current y-axis label"""
    assert isinstance(left, str) and isinstance(right, str)
    axe = gca()
    axe.ylabel = (left, right)


def zlabel(label):
    """Set current z-axis label"""
    assert isinstance(label, str)
    axe = gca()
    axe.zlabel = label


def yreverse(reverse):
    """
    Set y-axis direction of increasing values

    reverse = False (default)
        y-axis values increase from bottom to top

    reverse = True
        y-axis values increase from top to bottom
    """
    assert isinstance(reverse, bool)
    axe = gca()
    axe.yreverse = reverse


def grid(act):
    """Toggle grid visibility"""
    axe = gca()
    axe.set_grid(act)


def legend(pos="TR"):
    """Add legend to current axes (pos='TR', 'TL', 'BR', ...)"""
    axe = gca()
    axe.add_legend(pos)


def colormap(name):
    """Set color map to *name*"""
    axe = gca()
    axe.colormap = name


def _add_colormaps(glbs):
    """Add colormap functions to *glbs*"""
    for cmap_name in ALL_COLORMAPS.keys():
        glbs[cmap_name] = lambda name=cmap_name: colormap(name)
        glbs[cmap_name].__doc__ = "Set color map to '%s'" % cmap_name


_add_colormaps(globals())


def close(N=None, all=False):
    """Close figure"""
    global _figures, _current_fig, _current_axes
    if all:
        _figures = {}
        _current_fig = None
        _current_axes = None
        return
    if N is None:
        fig = gcf()
    else:
        fig = figure(N)
    fig.close()


def savefig(fname, format=None):
    """
    Save figure

    Currently supports QImageWriter formats only
    (see https://doc.qt.io/qt-5/qimagewriter.html#supportedImageFormats)
    """
    if not isinstance(fname, str) and format is None:
        # Buffer/fd
        format = "png"
    if format is None:
        format = fname.rsplit(".", 1)[-1].lower()
        fmts = [fmt.data().decode() for fmt in QG.QImageWriter.supportedImageFormats()]
        assert format in fmts, _(
            "Function 'savefig' currently supports the " "following formats:\n%s"
        ) % ", ".join(fmts)
    else:
        format = format.lower()
    fig = gcf()
    fig.save(fname, format)



# -*- coding: utf-8 -*-

from __future__ import annotations

from copy import deepcopy
from math import atan2, pi, sqrt
from typing import TYPE_CHECKING

from guidata.dataset import (
    BeginGroup,
    BoolItem,
    ButtonItem,
    ColorItem,
    DataSet,
    DataSetGroup,
    EndGroup,
    FloatItem,
    ImageChoiceItem,
    IntItem,
    ObjectItem,
    StringItem,
    update_dataset,
)
from guidata.dataset.qtitemwidgets import DataSetWidget
from guidata.dataset.qtwidgets import DataSetEditLayout
from qtpy import QtCore as QC
from qtpy import QtGui as QG
from qtpy import QtWidgets as QW
from qwt import QwtPlotCurve, QwtPlotMarker, QwtSymbol

from plotpy.config import _

if TYPE_CHECKING:
    import guidata.dataset

    from plotpy.interfaces import IBasePlotItem
    from plotpy.plot import BasePlot

LINESTYLES = {"-": "SolidLine", "--": "DashLine", ":": "DotLine", "-.": "DashDotLine"}
COLORS = {
    "r": "red",
    "g": "green",
    "b": "blue",
    "c": "cyan",
    "m": "magenta",
    "y": "yellow",
    "k": "black",
    "w": "white",
    "G": "gray",
}
MARKERS = {
    "+": QwtSymbol.Cross,
    "o": QwtSymbol.Ellipse,
    "*": QwtSymbol.Star1,
    ".": QwtSymbol(
        QwtSymbol.Ellipse, QG.QBrush(QC.Qt.black), QG.QPen(QC.Qt.black), QC.QSizeF(3, 3)
    ),
    "x": QwtSymbol.XCross,
    "s": QwtSymbol.Rect,
    "d": QwtSymbol.Diamond,
    "^": QwtSymbol.UTriangle,
    "v": QwtSymbol.DTriangle,
    ">": QwtSymbol.RTriangle,
    "<": QwtSymbol.LTriangle,
    "h": QwtSymbol.Star2,
}
MARKERSTYLES = {None: "NoLine", "-": "HLine", "|": "VLine", "+": "Cross"}


LINESTYLE_CHOICES = [
    ("SolidLine", _("Solid line"), "solid.png"),
    ("DashLine", _("Dashed line"), "dash.png"),
    ("DotLine", _("Dotted line"), "dot.png"),
    ("DashDotLine", _("Dash-dot line"), "dashdot.png"),
    ("DashDotDotLine", _("Dash-dot-dot line"), "dashdotdot.png"),
    ("NoPen", _("No line"), "none.png"),
]
MARKER_CHOICES = [
    ("Cross", _("Cross"), "cross.png"),
    ("Ellipse", _("Ellipse"), "ellipse.png"),
    ("Star1", _("Star"), "star.png"),
    ("XCross", _("X-Cross"), "xcross.png"),
    ("Rect", _("Square"), "square.png"),
    ("Diamond", _("Diamond"), "diamond.png"),
    ("UTriangle", _("Triangle"), "triangle_u.png"),
    ("DTriangle", _("Triangle"), "triangle_d.png"),
    ("RTriangle", _("Triangle"), "triangle_r.png"),
    ("LTriangle", _("Triangle"), "triangle_l.png"),
    ("Star2", _("Hexagon"), "hexagon.png"),
    ("NoSymbol", _("No symbol"), "none.png"),
]
CURVESTYLE_CHOICES = [
    ("Lines", _("Lines"), "lines.png"),
    ("Sticks", _("Sticks"), "sticks.png"),
    ("Steps", _("Steps"), "steps.png"),
    ("Dots", _("Dots"), "dots.png"),
    ("NoCurve", _("No curve"), "none.png"),
]

BRUSHSTYLE_CHOICES = [
    ("NoBrush", _("No brush pattern"), "nobrush.png"),
    ("SolidPattern", _("Uniform color"), "solidpattern.png"),
    ("Dense1Pattern", _("Extremely dense brush pattern"), "dense1pattern.png"),
    ("Dense2Pattern", _("Very dense brush pattern"), "dense2pattern.png"),
    ("Dense3Pattern", _("Somewhat dense brush pattern"), "dense3pattern.png"),
    ("Dense4Pattern", _("Half dense brush pattern"), "dense4pattern.png"),
    ("Dense5Pattern", _("Somewhat sparse brush pattern"), "dense5pattern.png"),
    ("Dense6Pattern", _("Very sparse brush pattern"), "dense6pattern.png"),
    ("Dense7Pattern", _("Extremely sparse brush pattern"), "dense7pattern.png"),
    ("HorPattern", _("Horizontal lines"), "horpattern.png"),
    ("VerPattern", _("Vertical lines"), "verpattern.png"),
    ("CrossPattern", _("Crossing horizontal and vertical lines"), "crosspattern.png"),
    ("BDiagPattern", _("Backward diagonal lines"), "bdiagpattern.png"),
    ("FDiagPattern", _("Forward diagonal lines"), "fdiagpattern.png"),
    ("DiagCrossPattern", _("Crossing diagonal lines"), "diagcrosspattern.png"),
    # (
    #     "LinearGradientPattern",
    #     _("Linear gradient (set using a dedicated QBrush constructor)"),
    #     "none.png",
    # ),
    # (
    #     "ConicalGradientPattern",
    #     _("Conical gradient (set using a dedicated QBrush constructor)"),
    #     "none.png",
    # ),
    # (
    #     "RadialGradientPattern",
    #     _("Radial gradient (set using a dedicated QBrush constructor)"),
    #     "none.png",
    # ),
    # ("TexturePattern", _("Custom pattern (see QBrush::setTexture())"), "none.png"),
]

MARKERSTYLE_CHOICES = [
    ("NoLine", _("None"), "none.png"),
    ("HLine", _("Horizontal"), "horiz_marker.png"),
    ("VLine", _("Vertical"), "vert_marker.png"),
    ("Cross", _("Cross"), "cross_marker.png"),
]


def build_reverse_map(lst, obj):
    """

    :param lst:
    :param obj:
    :return:
    """
    dict = {}
    for idx, _name, _icon in lst:
        val = getattr(obj, idx)
        dict[val] = idx
    return dict


MARKER_NAME = build_reverse_map(MARKER_CHOICES, QwtSymbol)
CURVESTYLE_NAME = build_reverse_map(CURVESTYLE_CHOICES, QwtPlotCurve)
LINESTYLE_NAME = build_reverse_map(LINESTYLE_CHOICES, QC.Qt)
BRUSHSTYLE_NAME = build_reverse_map(BRUSHSTYLE_CHOICES, QC.Qt)
MARKERSTYLE_NAME = build_reverse_map(MARKERSTYLE_CHOICES, QwtPlotMarker)


def style_generator(color_keys="bgrcmykG"):
    """Cycling through curve styles"""
    while True:
        for linestyle in sorted(LINESTYLES.keys()):
            for color in color_keys:
                yield color + linestyle


def update_style_attr(style, param):
    """Parse a MATLAB-like style string and
    update the color, linestyle, marker attributes of the param
    object
    """
    for marker in list(MARKERS.keys()):
        if marker in style:
            param.symbol.update_param(MARKERS[marker])
            break
    else:
        param.symbol.update_param(QwtSymbol.NoSymbol)
    for linestyle in list(LINESTYLES.keys()):
        if linestyle in style:
            param.line.style = LINESTYLES[linestyle]
            break
    else:
        param.line.style = "NoPen"
    for color in list(COLORS.keys()):
        if color in style:
            param.line.color = COLORS[color]
            param.symbol.facecolor = COLORS[color]
            param.symbol.edgecolor = COLORS[color]
            break


class ItemParameters:
    """Class handling QwtPlotItem-like parameters

    Args:
        multiselection (bool): if True, the class will handle
    """

    MULTISEL_DATASETS = []
    # Customizing tab display order:
    ENDING_PARAMETERS = (
        "CurveParam",
        "ErrorBarParam",
        "ShapeParam",
        "LabelParam",
        "LegendParam",
        "GridParam",
        "AxesParam",
    )

    def __init__(self, multiselection: bool = False):
        self.multiselection = multiselection
        self.paramdict: dict[str, guidata.dataset.DataSet] = {}
        self.items: set[IBasePlotItem] = set()

    @classmethod
    def register_multiselection(cls, klass, klass_ms):
        """Register a DataSet couple: (DataSet, DataSet_for_MultiSelection)"""
        # Inserting element backwards because classes have to be registered
        # from children to parent (see 'add' method to fully understand why)
        cls.MULTISEL_DATASETS.insert(0, (klass, klass_ms))

    def __add(self, key, item, param):
        self.paramdict[key] = param
        self.items.add(item)

    def add(
        self,
        key: str,
        item: IBasePlotItem,
        param: guidata.dataset.DataSet,
    ) -> None:
        """
        Add parameters for a given item

        Args:
            key: key to identify the item
            item: item to be customized
            param: parameters to be applied to the item
        """
        if self.multiselection:
            for klass, klass_ms in self.MULTISEL_DATASETS:
                if isinstance(param, klass):
                    title = param.get_title()
                    if key in self.paramdict and not title.endswith("s"):
                        title += "s"
                    param_ms = klass_ms(
                        title=title, comment=param.get_comment(), icon=param.get_icon()
                    )
                    update_dataset(param_ms, param)
                    self.__add(key, item, param_ms)
                    return
        self.__add(key, item, param)

    def get(self, key: str) -> guidata.dataset.DataSet:
        """
        Get parameters for a given item

        Args:
            key: key to identify the item
        """
        return deepcopy(self.paramdict.get(key))

    def update(self, plot: BasePlot) -> None:
        """
        Update plot items according to the parameters

        Args:
            plot: plot to be updated
        """
        for item in self.items:
            item.set_item_parameters(self)
            plot.SIG_ITEM_PARAMETERS_CHANGED.emit(item)
        plot.replot()
        plot.SIG_ITEMS_CHANGED.emit(plot)

    def edit(self, plot: BasePlot, title: str, icon: str) -> None:
        """
        Edit parameters

        Args:
            plot: plot to be updated
            title: title of the dialog
            icon: icon of the dialog
        """
        paramdict = self.paramdict.copy()
        ending_parameters = []
        for key in self.ENDING_PARAMETERS:
            if key in paramdict:
                ending_parameters.append(paramdict.pop(key))
        parameters = list(paramdict.values()) + ending_parameters
        dset = DataSetGroup(parameters, title=title.rstrip("."), icon=icon)
        if dset.edit(parent=plot, apply=lambda dset: self.update(plot)):
            self.update(plot)


# ===================================================
# Common font parameters
# ===================================================
def _font_selection(param, item, value, parent):
    font = param.build_font()
    result, valid = QW.QFontDialog.getFont(font, parent)
    if valid:
        param.update_param(result)


class FontParam(DataSet):
    family = StringItem(_("Family"), default="default")
    _choose = ButtonItem(_("Choose font"), _font_selection, default=None).set_pos(col=1)
    size = IntItem(_("Size in point"), default=12)
    bold = BoolItem(_("Bold"), default=False).set_pos(col=1)
    italic = BoolItem(_("Italic"), default=False).set_pos(col=2)

    def update_param(self, font):
        """

        :param font:
        """
        self.family = str(font.family())
        self.size = font.pointSize()
        self.bold = bool(font.bold())
        self.italic = bool(font.italic())

    def build_font(self):
        """

        :return:
        """
        font = QG.QFont(self.family)
        font.setPointSize(self.size)
        font.setBold(self.bold)
        font.setItalic(self.italic)
        return font


class FontItemWidget(DataSetWidget):
    klass = FontParam


class FontItem(ObjectItem):
    """Item holding a LineStyleParam"""

    klass = FontParam


DataSetEditLayout.register(FontItem, FontItemWidget)


# ===================================================
# Common Qwt symbol parameters
# ===================================================
class SymbolParam(DataSet):
    marker = ImageChoiceItem(_("Style"), MARKER_CHOICES, default="NoSymbol")
    size = IntItem(_("Size"), default=9)
    edgecolor = ColorItem(_("Border"), default="gray")
    facecolor = ColorItem(_("Background color"), default="yellow")
    alpha = FloatItem(_("Background alpha"), default=1.0, min=0, max=1)

    def update_param(self, symb):
        """

        :param symb:
        :return:
        """
        if not isinstance(symb, QwtSymbol):
            # check if this is still needed
            # raise RuntimeError
            assert isinstance(symb, QwtSymbol.Style)
            self.marker = MARKER_NAME[symb]
            return
        self.marker = MARKER_NAME[symb.style()]
        self.size = symb.size().width()
        self.edgecolor = str(symb.pen().color().name())
        self.facecolor = str(symb.brush().color().name())

    def build_symbol(self):
        """

        :return:
        """
        marker_type = getattr(QwtSymbol, self.marker)
        color = QG.QColor(self.facecolor)
        color.setAlphaF(self.alpha)
        marker = QwtSymbol(
            marker_type,
            QG.QBrush(color),
            QG.QPen(QG.QColor(self.edgecolor)),
            QC.QSizeF(self.size, self.size),
        )
        return marker

    def update_symbol(self, obj):
        """

        :param obj:
        """
        obj.setSymbol(self.build_symbol())


class SymbolItemWidget(DataSetWidget):
    klass = SymbolParam


class SymbolItem(ObjectItem):
    """Item holding a SymbolParam"""

    klass = SymbolParam


DataSetEditLayout.register(SymbolItem, SymbolItemWidget)


# ===================================================
# Common line style parameters
# ===================================================
class LineStyleParam(DataSet):
    style = ImageChoiceItem(_("Style"), LINESTYLE_CHOICES, default="SolidLine")
    color = ColorItem(_("Color"), default="black")
    width = FloatItem(_("Width"), default=1.0, min=0)

    def update_param(self, pen):
        """

        :param pen:
        """
        self.width = pen.widthF()
        self.color = str(pen.color().name())
        self.style = LINESTYLE_NAME[pen.style()]

    def build_pen(self):
        """

        :return:
        """
        linecolor = QG.QColor(self.color)
        style = getattr(QC.Qt, self.style)
        pen = QG.QPen(linecolor, self.width, style)
        return pen

    def set_style_from_matlab(self, linestyle):
        """Eventually convert MATLAB-like linestyle into Qt linestyle"""
        linestyle = LINESTYLES.get(linestyle, linestyle)  # MATLAB-style
        if linestyle == "":  # MATLAB-style
            linestyle = "NoPen"
        self.style = linestyle


class LineStyleItemWidget(DataSetWidget):
    klass = LineStyleParam


class LineStyleItem(ObjectItem):
    """Item holding a LineStyleParam"""

    klass = LineStyleParam


DataSetEditLayout.register(LineStyleItem, LineStyleItemWidget)


# ===================================================
# Common brush style parameters
# ===================================================
class BrushStyleParam(DataSet):
    style = ImageChoiceItem(_("Style"), BRUSHSTYLE_CHOICES, default="SolidPattern")
    color = ColorItem(_("Color"), default="black")
    alpha = FloatItem(_("Alpha"), default=1.0)
    angle = FloatItem(_("Angle"), default=0.0, min=0)
    sx = FloatItem(_("sx"), default=1.0, min=0)
    sy = FloatItem(_("sy"), default=1.0, min=0)

    def update_param(self, brush):
        """

        :param brush:
        """
        tr = brush.transform()
        pt = tr.map(QC.QPointF(1.0, 0.0))
        self.sx = sqrt(pt.x() ** 2 + pt.y() ** 2)
        self.angle = 180 * atan2(pt.y(), pt.x()) / pi
        pt = tr.map(QC.QPointF(0.0, 1.0))
        self.sy = sqrt(pt.x() ** 2 + pt.y() ** 2)

        col = brush.color()
        self.color = str(col.name())
        self.alpha = col.alphaF()
        self.style = BRUSHSTYLE_NAME[brush.style()]

    def build_brush(self):
        """

        :return:
        """
        color = QG.QColor(self.color)
        color.setAlphaF(self.alpha)
        brush = QG.QBrush(color, getattr(QC.Qt, self.style))
        tr = QG.QTransform()
        tr = tr.scale(self.sx, self.sy)
        tr = tr.rotate(self.angle)
        brush.setTransform(tr)
        return brush


class BrushStyleItemWidget(DataSetWidget):
    klass = BrushStyleParam


class BrushStyleItem(ObjectItem):
    """Item holding a LineStyleParam"""

    klass = BrushStyleParam


DataSetEditLayout.register(BrushStyleItem, BrushStyleItemWidget)


# ===================================================
# QwtText parameters
# ===================================================
class TextStyleParam(DataSet):
    font = FontItem(_("Font"))
    textcolor = ColorItem(_("Text color"), default="blue")
    background_color = ColorItem(_("Background color"), default="white")
    background_alpha = FloatItem(_("Background alpha"), default=0.5, min=0, max=1)

    def update_param(self, obj):
        """obj: QwtText instance"""
        self.font.update_param(obj.font())
        self.textcolor = obj.color().name()
        color = obj.backgroundBrush().color()
        self.background_color = color.name()
        self.background_alpha = color.alphaF()

    def update_text(self, obj):
        """obj: QwtText instance"""
        obj.setColor(QG.QColor(self.textcolor))
        color = QG.QColor(self.background_color)
        color.setAlphaF(self.background_alpha)
        obj.setBackgroundBrush(QG.QBrush(color))
        font = self.font.build_font()
        obj.setFont(font)


class TextStyleItemWidget(DataSetWidget):
    klass = TextStyleParam


class TextStyleItem(ObjectItem):
    """Item holding a TextStyleParam"""

    klass = TextStyleParam


DataSetEditLayout.register(TextStyleItem, TextStyleItemWidget)


# ===================================================
# Grid parameters
# ===================================================
class GridParam(DataSet):
    background = ColorItem(_("Background color"), default="white")
    maj = BeginGroup(_("Major grid"))
    maj_xenabled = BoolItem(_("X Axis"), default=True)
    maj_yenabled = BoolItem(_("Y Axis"), default=True).set_pos(col=1)
    maj_line = LineStyleItem(_("Line"))
    _maj = EndGroup("end group")

    min = BeginGroup(_("Minor grid"))
    min_xenabled = BoolItem(_("X Axis"), default=False)
    min_yenabled = BoolItem(_("Y Axis"), default=False).set_pos(col=1)
    min_line = LineStyleItem(_("Line"))
    _min = EndGroup("fin groupe")

    def update_param(self, grid):
        """

        :param grid:
        """
        plot = grid.plot()
        if plot is not None:
            self.background = str(plot.canvasBackground().color().name())
        self.maj_xenabled = grid.xEnabled()
        self.maj_yenabled = grid.yEnabled()
        self.maj_line.update_param(grid.majorPen())
        self.min_xenabled = grid.xMinEnabled()
        self.min_yenabled = grid.yMinEnabled()
        self.min_line.update_param(grid.minorPen())

    def update_grid(self, grid):
        """

        :param grid:
        """
        plot = grid.plot()
        if plot is not None:
            plot.blockSignals(True)  # Avoid unwanted calls of update_param
            # triggered by the setter methods below
            plot.setCanvasBackground(QG.QColor(self.background))
        grid.enableX(self.maj_xenabled)
        grid.enableY(self.maj_yenabled)
        grid.setPen(self.maj_line.build_pen())
        grid.enableXMin(self.min_xenabled)
        grid.enableYMin(self.min_yenabled)
        grid.setMinorPen(self.min_line.build_pen())
        grid.setTitle(self.get_title())
        if plot is not None:
            plot.blockSignals(False)



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""Oblique averaged cross section test"""

# guitest: show

from __future__ import annotations

from plotpy.panels.csection.csitem import ObliqueCrossSectionItem
from plotpy.panels.csection.cswidget import ObliqueCrossSection
from plotpy.tests.tools import test_cross_section_line
from plotpy.tools import ImageMaskTool, ObliqueCrossSectionTool, OCSPanelTool

# debug mode shows the ROI in the top-left corner of the image plot:
ObliqueCrossSectionItem.DEBUG = False


class OCSImageDialog(test_cross_section_line.BaseCSImageDialog):
    """Oblique averaged cross section test"""

    TOOLCLASSES = (ObliqueCrossSectionTool, OCSPanelTool, ImageMaskTool)
    PANELCLASS = ObliqueCrossSection


def test_cross_section_oblique():
    """Test cross section oblique"""
    test_cross_section_line.generic_cross_section_dialog(
        "Oblique averaged cross section test", OCSImageDialog
    )


if __name__ == "__main__":
    test_cross_section_oblique()



# -*- coding: utf-8 -*-

"""
Plot Item list
^^^^^^^^^^^^^^

The `plot item list` panel is a widget which displays the list of items attached to
the plot.

.. autoclass:: PlotItemList
"""

from guidata.configtools import get_icon, get_image_layout
from guidata.qthelpers import add_actions, create_action
from guidata.utils.misc import assert_interfaces_valid
from qtpy import QtCore as QC
from qtpy import QtWidgets as QW

from plotpy.config import _
from plotpy.constants import ID_ITEMLIST
from plotpy.interfaces import IPanel
from plotpy.panels.base import PanelWidget


class ItemListWidget(QW.QListWidget):
    """
    PlotItemList
    List of items attached to plot
    """

    def __init__(self, parent):
        super().__init__(parent)

        self.manager = None
        self.plot = None  # the default plot...
        self.items = []

        self.currentRowChanged.connect(self.current_row_changed)
        self.itemChanged.connect(self.item_changed)
        self.itemSelectionChanged.connect(self.refresh_actions)
        self.itemSelectionChanged.connect(self.selection_changed)

        self.setWordWrap(True)
        self.setMinimumWidth(140)
        self.setSelectionMode(QW.QListWidget.ExtendedSelection)

        # Setup context menu
        self.menu = QW.QMenu(self)
        self.menu_actions = self.setup_actions()
        self.refresh_actions()
        add_actions(self.menu, self.menu_actions)

    def register_panel(self, manager):
        """

        :param manager:
        """
        self.manager = manager

        for plot in self.manager.get_plots():
            plot.SIG_ITEMS_CHANGED.connect(self.items_changed)
            plot.SIG_ACTIVE_ITEM_CHANGED.connect(self.active_item_changed)
        self.plot = self.manager.get_plot()

    def contextMenuEvent(self, event):
        """Override Qt method"""
        self.refresh_actions()
        self.menu.popup(event.globalPos())

    def setup_actions(self):
        """

        :return:
        """
        self.movedown_ac = create_action(
            self,
            _("Move to back"),
            icon=get_icon("arrow_down.png"),
            triggered=lambda: self.move_item("down"),
        )
        self.moveup_ac = create_action(
            self,
            _("Move to front"),
            icon=get_icon("arrow_up.png"),
            triggered=lambda: self.move_item("up"),
        )
        settings_ac = create_action(
            self,
            _("Parameters..."),
            icon=get_icon("settings.png"),
            triggered=self.edit_plot_parameters,
        )
        self.remove_ac = create_action(
            self, _("Remove"), icon=get_icon("trash.png"), triggered=self.remove_item
        )
        return [self.moveup_ac, self.movedown_ac, None, settings_ac, self.remove_ac]

    def edit_plot_parameters(self):
        """ """
        self.plot.edit_plot_parameters("item")

    def __is_selection_contiguous(self):
        indexes = sorted([self.row(lw_item) for lw_item in self.selectedItems()])
        return len(indexes) <= 1 or list(range(indexes[0], indexes[-1] + 1)) == indexes

    def get_selected_items(self):
        """Return selected QwtPlot items

        .. warning::

            This is not the same as
            :py:meth:`.baseplot.BasePlot.get_selected_items`.
            Some items could appear in itemlist without being registered in
            plot widget items (in particular, some items could be selected in
            itemlist without being selected in plot widget)
        """
        return [self.items[self.row(lw_item)] for lw_item in self.selectedItems()]

    def refresh_actions(self):
        """ """
        is_selection = len(self.selectedItems()) > 0
        for action in self.menu_actions:
            if action is not None:
                action.setEnabled(is_selection)
        if is_selection:
            remove_state = True
            for item in self.get_selected_items():
                remove_state = remove_state and not item.is_readonly()
            self.remove_ac.setEnabled(remove_state)
            for action in [self.moveup_ac, self.movedown_ac]:
                action.setEnabled(self.__is_selection_contiguous())

    def __get_item_icon(self, item):
        icon = item.icon()
        if icon is None:
            return get_icon("not_found.png")
        else:
            return icon

    def items_changed(self, plot):
        """Plot items have changed"""
        active_plot = self.manager.get_active_plot()
        if active_plot is not plot:
            return
        self.plot = plot
        _block = self.blockSignals(True)
        active = plot.get_active_item()
        self.items = plot.get_public_items(z_sorted=True)
        self.clear()
        for item in self.items:
            title = item.title().text()
            lw_item = QW.QListWidgetItem(self.__get_item_icon(item), title, self)
            lw_item.setCheckState(
                QC.Qt.Checked if item.isVisible() else QC.Qt.Unchecked
            )
            lw_item.setSelected(item.selected)
            font = lw_item.font()
            if item is active:
                font.setItalic(True)
            else:
                font.setItalic(False)
            lw_item.setFont(font)
            self.addItem(lw_item)
        self.refresh_actions()
        self.blockSignals(_block)

    def active_item_changed(self, plot):
        """Plot items have changed"""
        active_plot = self.manager.get_active_plot()
        if active_plot is not plot:
            return
        self.plot = plot
        _block = self.blockSignals(True)
        active = plot.get_active_item()
        for item in self.items:
            title = item.title().text()
            lw_item = self.item(self.items.index(item))
            lw_item.setText(title)
            lw_item.setSelected(item.selected)
            font = lw_item.font()
            if item is active:
                font.setItalic(True)
            else:
                font.setItalic(False)
            lw_item.setFont(font)
        self.refresh_actions()
        self.blockSignals(_block)

    def current_row_changed(self, index):
        """QListWidget current row has changed"""
        if index == -1:
            return
        item = self.items[index]
        if not item.can_select():
            item = None
        if item is None:
            self.plot.replot()

    def selection_changed(self):
        """ """
        items = self.get_selected_items()
        self.plot.select_some_items(items)
        self.plot.replot()

    def item_changed(self, listwidgetitem):
        """QListWidget item has changed"""
        item = self.items[self.row(listwidgetitem)]
        visible = listwidgetitem.checkState() == QC.Qt.Checked
        if visible != item.isVisible():
            self.plot.set_item_visible(item, visible)

    def move_item(self, direction):
        """Move item to the background/foreground
        Works only for contiguous selection
        -> 'refresh_actions' method should guarantee that"""
        items = self.get_selected_items()
        if direction == "up":
            self.plot.move_up(items)
        else:
            self.plot.move_down(items)
        # Re-select items which can't be selected in plot widget but can be
        # selected in ItemListWidget:
        for item in items:
            lw_item = self.item(self.items.index(item))
            if not lw_item.isSelected():
                lw_item.setSelected(True)
        self.plot.replot()

    def remove_item(self):
        """ """
        if len(self.selectedItems()) == 1:
            message = _("Do you really want to remove this item?")
        else:
            message = _("Do you really want to remove selected items?")
        answer = QW.QMessageBox.warning(
            self, _("Remove"), message, QW.QMessageBox.Yes | QW.QMessageBox.No
        )
        if answer == QW.QMessageBox.Yes:
            items = self.get_selected_items()
            self.plot.del_items(items)
            self.plot.replot()


class PlotItemList(PanelWidget):
    """Construct the `plot item list panel`"""

    __implements__ = (IPanel,)
    PANEL_ID = ID_ITEMLIST
    PANEL_TITLE = _("Item list")
    PANEL_ICON = "item_list.png"

    def __init__(self, parent):
        super().__init__(parent)
        self.manager = None

        vlayout = QW.QVBoxLayout()
        self.setLayout(vlayout)

        style = "<span style='color: #444444'><b>%s</b></span>"
        layout, _label = get_image_layout(
            self.PANEL_ICON, style % self.PANEL_TITLE, alignment=QC.Qt.AlignCenter
        )
        vlayout.addLayout(layout)
        self.listwidget = ItemListWidget(self)
        vlayout.addWidget(self.listwidget)

        toolbar = QW.QToolBar(self)
        vlayout.addWidget(toolbar)
        add_actions(toolbar, self.listwidget.menu_actions)

    def register_panel(self, manager):
        """Register panel to plot manager"""
        self.manager = manager
        self.listwidget.register_panel(manager)

    def configure_panel(self):
        """Configure panel"""
        pass


assert_interfaces_valid(PlotItemList)



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""
plotpy.widgets.qtdesigner
-----------------------------

The `qtdesigner` module provides QtDesigner helper functions for :mod:`plotpy`:
    * :py:func:`.qtdesigner.loadui`
    * :py:func:`.qtdesigner.compileui`
    * :py:func:`.qtdesigner.create_qtdesigner_plugins`

Reference
~~~~~~~~~

.. autofunction:: loadui
.. autofunction:: compileui
.. autofunction:: create_qtdesigner_plugin
"""

import io

from guidata.configtools import get_icon
from qtpy import QtGui as QG
from qtpy import uic
from qtpy.QtDesigner import QPyDesignerCustomWidgetPlugin


def loadui(fname, replace_class="QwtPlot"):
    """
    Return Widget or Window class from QtDesigner ui file 'fname'

    The loadUiType function (PyQt5.uic) doesn't work correctly with plotpy
    QtDesigner plugins because they don't inheritate from a PyQt5.QtGui
    object.
    """
    with open(fname) as f:
        uifile_text = f.read().replace(replace_class, "QFrame")
    ui, base_class = uic.loadUiType(io.StringIO(uifile_text))

    class Form(base_class, ui):
        """ """

        def __init__(self, parent=None):
            super().__init__(parent)
            self.setupUi(self)

    return Form


def compileui(fname, replace_class="QwtPlot"):
    """

    :param fname:
    :param replace_class:
    """
    with open(fname) as f:
        uifile_text = f.read().replace("QwtPlot", "QFrame")
    with open(fname.replace(".ui", "_ui.py"), "w") as pyfile:
        uic.compileUi(io.StringIO(uifile_text), pyfile, pyqt3_wrapper=True)


def create_qtdesigner_plugin(
    group,
    module_name,
    class_name,
    plot_options={},
    icon=None,
    tooltip="",
    whatsthis="",
):
    """Return a custom QtDesigner plugin class

    Example:
    create_qtdesigner_plugin(group = "plotpy", module_name = "plotpy.baseplot",
                             class_name = "PlotWidget",
                             plot_options=PlotOptions(type="image")),
                             icon = "image.png", tooltip = "", whatsthis = ""):
    """
    Widget = getattr(__import__(module_name, fromlist=[class_name]), class_name)

    class CustomWidgetPlugin(QPyDesignerCustomWidgetPlugin):
        """ """

        def __init__(self, parent=None):
            QPyDesignerCustomWidgetPlugin.__init__(self)
            self.initialized = False

        def initialize(self, core):
            """

            :param core:
            :return:
            """
            if self.initialized:
                return
            self.initialized = True

        def isInitialized(self):
            """

            :return:
            """
            return self.initialized

        def createWidget(self, parent):
            """

            :param parent:
            :return:
            """
            return Widget(parent, options=plot_options)

        def name(self):
            """

            :return:
            """
            return class_name

        def group(self):
            """

            :return:
            """
            return group

        def icon(self):
            """

            :return:
            """
            if icon is not None:
                return get_icon(icon)
            else:
                return QG.QIcon()

        def toolTip(self):
            """

            :return:
            """
            return tooltip

        def whatsThis(self):
            """

            :return:
            """
            return whatsthis

        def isContainer(self):
            """

            :return:
            """
            return False

        def domXml(self):
            """

            :return:
            """
            return f'<widget class="{class_name}" name="{class_name.lower()}" />\n'

        def includeFile(self):
            """

            :return:
            """
            return module_name

    return CustomWidgetPlugin



# -*- coding: utf-8 -*-

"""Update requirements.rst file from pyproject.toml or setup.cfg file

Warning: this has to be done manually at release time.
It is not done automatically by the sphinx 'conf.py' file because it
requires an internet connection to fetch the dependencies metadata - this
is not always possible (e.g., when building the documentation on a machine
without internet connection like the Debian package management infrastructure).
"""

from guidata.utils.genreqs import gen_module_req_rst  # noqa: E402

import plotpy

if __name__ == "__main__":
    print("Updating requirements.rst file...", end=" ")
    gen_module_req_rst(plotpy, ["Python>=3.9", "PyQt5>=5.11"])
    print("done.")



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

# pylint: disable=C0103

"""
Item builder
------------

The `builder` module provides a builder singleton class that can be
used to simplify the creation of plot items.
"""

from __future__ import annotations

from .annotation import AnnotationBuilder
from .curvemarker import CurveMarkerCursorBuilder
from .image import ImageBuilder
from .label import LabelBuilder
from .plot import WidgetBuilder
from .shape import ShapeBuilder


class PlotBuilder(
    WidgetBuilder,
    CurveMarkerCursorBuilder,
    ImageBuilder,
    LabelBuilder,
    ShapeBuilder,
    AnnotationBuilder,
):
    """Class regrouping a set of factory functions to simplify the creation
    of plot widgets and plot items.

    It is a singleton class, so you should not create instances of this class
    but use the :py:data:`plotpy.builder.make` instance instead.
    """

    def __init__(self):
        super().__init__()


#: Instance of :py:class:`.PlotBuilder`
make = PlotBuilder()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""Contrast tool test"""

# guitest: show

import os
import os.path as osp

from guidata.env import execenv
from guidata.qthelpers import qt_app_context

from plotpy.builder import make
from plotpy.tests import get_path
from plotpy.tests.data import gen_image1


def __create_dialog_with_contrast(item):
    """Create plot dialog with contrast panel

    Args:
        Item: item to be added to the plot
    """
    win = make.dialog(
        edit=False,
        toolbar=True,
        wintitle="Contrast test",
        show_contrast=True,
        type="image",
        size=(600, 600),
    )
    plot = win.get_plot()
    plot.add_item(item)
    plot.set_active_item(item)
    item.unselect()
    win.show()
    return win


def test_contrast1():
    """Contrast test 1"""
    with qt_app_context(exec_loop=True):
        item = make.image(filename=get_path("brain.png"), colormap="gray")
        win = __create_dialog_with_contrast(item)
        fname = "contrast.png"
        try:
            win.get_plot().save_widget(fname)
        except IOError:
            # Skipping this part of the test
            # because user has no write permission on current directory
            pass
        if execenv.unattended and osp.isfile(fname):
            os.unlink(fname)


def test_contrast2():
    """Contrast test 2

    Test if level histogram is really removed when the associated image is removed from
    the plot (the validation is not automatic)
    """
    with qt_app_context(exec_loop=True):
        item1 = make.image(filename=get_path("brain.png"), colormap="gray")
        win = __create_dialog_with_contrast(item1)
        plot = win.get_plot()
        plot.del_item(item1)
        item2 = make.image(gen_image1())
        plot.add_item(item2)
        plot.set_active_item(item2)


if __name__ == "__main__":
    # test_contrast1()
    test_contrast2()



# -*- coding: utf-8 -*-
#
# For licensing and distribution details, please read carefully xgrid/__init__.py

"""
Get rectangular selection from image
"""

# guitest: show

import numpy as np
from guidata.env import execenv
from guidata.qthelpers import qt_app_context

from plotpy.builder import make
from plotpy.tests.tools.test_get_segment import SEG_AXES_COORDS, PatchedSelectDialog
from plotpy.tools import RectangleTool
from plotpy.widgets.selectdialog import select_with_shape_tool


def test_get_rectangle():
    """Test get_rectangle"""
    with qt_app_context():
        image = make.image(data=np.random.rand(200, 200), colormap="gray")
        shape = select_with_shape_tool(
            None, RectangleTool, image, "Test", tooldialogclass=PatchedSelectDialog
        )
        if shape is not None:
            rect = shape.get_rect()
            if execenv.unattended:
                assert [round(i) for i in list(rect)] == SEG_AXES_COORDS
            elif rect is not None:
                print("Area:", rect)


if __name__ == "__main__":
    test_get_rectangle()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""
SelectPointTool test

This plotpy tool provide a MATLAB-like "ginput" feature.
"""

# guitest: show

from __future__ import annotations

from guidata.qthelpers import exec_dialog, qt_app_context
from numpy import linspace, sin

from plotpy.builder import make
from plotpy.config import _
from plotpy.tools import SelectPointsTool


def callback_function(tool: SelectPointsTool) -> None:
    """Callback function to be called by the tool after selection

    Args:
        tool: The tool instance that called the callback
    """
    points = tool.get_coordinates()
    print("Points : ", points)


def get_points(cdata: tuple[tuple[float, float], ...], max_select: int) -> None:
    """
    Plot curves and return selected point(s) coordinates

    Args:
        cdata: A tuple of curves to plot
    """
    win = make.dialog(
        wintitle=_(
            "Select up to %s points then press OK to accept "
            "(hold Ctrl to select multiple points)"
        )
        % max_select,
        edit=True,
        type="curve",
        curve_antialiasing=True,
        size=(800, 600),
    )
    default = win.manager.add_tool(
        SelectPointsTool,
        title="Test",
        on_active_item=True,
        mode="reuse",
        end_callback=callback_function,
        max_select=max_select,
    )
    default.activate()
    plot = win.manager.get_plot()
    for cx, cy in cdata[:-1]:
        item = make.mcurve(cx, cy)
        plot.add_item(item)
    # the last curve will be actibe and needs to have a different style for explictness
    item = make.mcurve(*cdata[-1], "r-+")
    plot.add_item(item)
    plot.set_active_item(item)
    plot.unselect_item(item)
    exec_dialog(win)


def test_get_point():
    """Test"""
    with qt_app_context(exec_loop=False):
        x = linspace(-10, 10, num=200)
        y = 0.25 * sin(sin(sin(x * 0.5)))
        x2 = linspace(-10, 10, 200)
        y2 = sin(sin(sin(x2)))
        get_points(((x, y), (x2, y2), (x, sin(2 * y))), max_select=5)


if __name__ == "__main__":
    test_get_point()



# -*- coding: utf-8 -*-

# pylint: disable=C0103

"""
Histogram item
"""

from __future__ import annotations

import weakref
from typing import TYPE_CHECKING

import numpy as np
from guidata.configtools import get_icon
from guidata.dataset import update_dataset
from guidata.utils.misc import assert_interfaces_valid
from qwt import QwtPlotCurve

from plotpy.config import _
from plotpy.interfaces import IBasePlotItem, IHistDataSource
from plotpy.items.curve.base import CurveItem
from plotpy.styles.curve import CurveParam
from plotpy.styles.histogram import HistogramParam

if TYPE_CHECKING:
    from plotpy.items.image.base import BaseImageItem
    from plotpy.styles.base import ItemParameters


class HistDataSource:
    """An objects that provides an Histogram data source interface
    to a simple numpy array of data
    """

    __implements__ = (IHistDataSource,)

    def __init__(self, data: np.ndarray) -> None:
        self.data = data

    def get_histogram(
        self, nbins: int, drange: tuple[float, float] | None = None
    ) -> tuple[np.ndarray, np.ndarray]:
        """
        Return a tuple (hist, bins) where hist is a list of histogram values

        Args:
            nbins: number of bins
            drange: lower and upper range of the bins. If not provided, range is
             simply (data.min(), data.max()). Values outside the range are ignored.

        Returns:
            Tuple (hist, bins)
        """
        return np.histogram(self.data, bins=nbins, range=drange)


assert_interfaces_valid(HistDataSource)


class HistogramItem(CurveItem):
    """Histogram plot item

    Args:
        curveparam: Curve parameters
        histparam: Histogram parameters
    """

    __implements__ = (IBasePlotItem,)

    def __init__(
        self,
        curveparam: CurveParam | None = None,
        histparam: HistogramParam | None = None,
        keep_weakref: bool = False,
    ) -> None:
        self.hist_count = None
        self.hist_bins = None
        self.bins = None
        self.bin_range = None
        self.old_bins = None
        self.source: BaseImageItem | None = None
        self.logscale: bool | None = None
        self.old_logscale = None
        self.keep_weakref = keep_weakref
        if curveparam is None:
            curveparam = CurveParam(_("Curve"), icon="curve.png")
            curveparam.curvestyle = "Steps"
        if histparam is None:
            self.histparam = HistogramParam(title=_("Histogram"), icon="histogram.png")
        else:
            self.histparam = histparam
        CurveItem.__init__(self, curveparam)
        self.setCurveAttribute(QwtPlotCurve.Inverted)
        self.setIcon(get_icon("histogram.png"))

    def set_hist_source(self, src: BaseImageItem) -> None:
        """Set histogram source

        Args:
            src: Object with method `get_histogram`, e.g. objects derived from
             :py:class:`.ImageItem`
        """
        if self.keep_weakref:
            self.source = weakref.ref(src)
        else:
            self.source = src
        self.update_histogram()

    def get_hist_source(self) -> BaseImageItem | None:
        """Return histogram source

        Returns:
            object: Object with method `get_histogram`, e.g. objects derived from
             :py:class:`.ImageItem`
        """
        if self.source is not None:
            if self.keep_weakref:
                return self.source()
            return self.source

    def set_hist_data(self, data: np.ndarray) -> None:
        """Set histogram data

        Args:
            data: numpy array
        """
        self.set_hist_source(HistDataSource(data))

    def set_logscale(self, state: bool) -> None:
        """Sets whether we use a logarithm or linear scale
        for the histogram counts

        Args:
            state: True for logarithmic scale
        """
        self.logscale = state
        self.update_histogram()

    def get_logscale(self) -> bool | None:
        """Returns the status of the scale

        Returns:
            bool: True for logarithmic scale
        """
        return self.logscale

    def set_bins(self, n_bins: int) -> None:
        """Sets the number of bins

        Args:
            n_bins: number of bins
        """
        self.bins = n_bins
        self.update_histogram()

    def get_bins(self) -> int | None:
        """Returns the number of bins

        Returns:
            int: number of bins
        """
        return self.bins

    def set_bin_range(self, bin_range: tuple[float, float] | None) -> None:
        """Sets the range of the bins

        Args:
            bin_range: (min, max) or None for automatic range
        """
        self.bin_range = bin_range
        self.update_histogram()

    def get_bin_range(self) -> tuple[float, float] | None:
        """Returns the range of the bins

        Returns:
            tuple: (min, max)
        """
        return self.bin_range

    def compute_histogram(self) -> tuple[np.ndarray, np.ndarray]:
        """Compute histogram data

        Returns:
            tuple: (hist, bins)
        """
        return self.get_hist_source().get_histogram(self.bins, self.bin_range)

    def update_histogram(self) -> None:
        """Update histogram data"""
        if self.get_hist_source() is None:
            return
        hist, bin_edges = self.compute_histogram()
        # Duplicate the first `hist` value to get a step-like histogram:
        hist = np.concatenate(([hist[0]], hist))
        if self.logscale:
            hist = np.log(hist + 1)

        self.set_data(bin_edges, hist)
        # Autoscale only if logscale/bins have changed
        if self.bins != self.old_bins or self.logscale != self.old_logscale:
            if self.plot():
                self.plot().do_autoscale()
        self.old_bins = self.bins
        self.old_logscale = self.logscale

        plot = self.plot()
        if plot is not None:
            plot.do_autoscale(replot=True)

    def update_params(self):
        """Update histogram parameters"""
        self.histparam.update_hist(self)
        CurveItem.update_params(self)

    def get_item_parameters(self, itemparams: ItemParameters) -> None:
        """
        Appends datasets to the list of DataSets describing the parameters
        used to customize apearance of this item

        Args:
            itemparams: Item parameters
        """
        CurveItem.get_item_parameters(self, itemparams)
        itemparams.add("HistogramParam", self, self.histparam)

    def set_item_parameters(self, itemparams):
        """
        :param itemparams:
        """
        update_dataset(
            self.histparam, itemparams.get("HistogramParam"), visible_only=True
        )
        self.histparam.update_hist(self)
        CurveItem.set_item_parameters(self, itemparams)


assert_interfaces_valid(HistogramItem)



# -*- coding: utf-8 -*-

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from guidata.configtools import get_icon
from guidata.utils.misc import assert_interfaces_valid

from plotpy.items.shape.polygon import PolygonShape

if TYPE_CHECKING:
    from plotpy.styles.shape import ShapeParam


class SegmentShape(PolygonShape):
    """Segment shape

    Args:
        x1: X coordinate of the first point
        y1: Y coordinate of the first point
        x2: X coordinate of the second point
        y2: Y coordinate of the second point
        shapeparam: Shape parameters
    """

    CLOSED = False
    ADDITIONNAL_POINTS = 1  # Number of points which are not part of the shape

    def __init__(
        self,
        x1: float = 0.0,
        y1: float = 0.0,
        x2: float = 0.0,
        y2: float = 0.0,
        shapeparam: ShapeParam = None,
    ):
        super().__init__(shapeparam=shapeparam)
        self.set_rect(x1, y1, x2, y2)
        self.setIcon(get_icon("segment.png"))

    def set_rect(self, x1: float, y1: float, x2: float, y2: float) -> None:
        """Set the segment coordinates

        Args:
            x1: X coordinate of the first point
            y1: Y coordinate of the first point
            x2: X coordinate of the second point
            y2: Y coordinate of the second point
        """
        self.set_points([(x1, y1), (x2, y2), (0.5 * (x1 + x2), 0.5 * (y1 + y2))])

    def get_rect(self) -> tuple[float, float, float, float]:
        """Return the segment coordinates

        Returns:
            Segment coordinates as a tuple (x1, y1, x2, y2)
        """
        return tuple(self.points[0]) + tuple(self.points[1])

    def move_point_to(
        self, handle: int, pos: tuple[float, float], ctrl: bool = False
    ) -> None:
        """Move a handle as returned by hit_test to the new position

        Args:
            handle: Handle
            pos: Position
            ctrl: True if <Ctrl> button is being pressed, False otherwise
        """
        nx, ny = pos
        x1, y1, x2, y2 = self.get_rect()
        if ctrl:
            # compute linear coefficient y = a * x + b
            # When ctrl is pressed, the point is moved on the line
            # defined by the segment, and the line is conserved
            # (the segment is not rotated)
            if x2 == x1:
                yA = ny
                xA = x1
            elif y2 == y1:
                yA = y1
                xA = nx
            else:
                a = (y2 - y1) / (x2 - x1)
                b = y1 - a * x1
                xA = a / (a * a + 1) * (nx / a + ny - b)
                yA = a * xA + b
            if handle == 0:
                self.set_rect(xA, yA, x2, y2)
            elif handle == 1:
                self.set_rect(x1, y1, xA, yA)
            elif handle in (2, -1):
                # selection du point milieu, on déplace le segment
                # c'est identique au comportmeent sans la touche Ctrl
                delta = (nx, ny) - self.points.mean(axis=0)
                self.points += delta
        else:
            # Ctrl is not pressed, the segment may be rotated freely
            if handle == 0:
                self.set_rect(nx, ny, x2, y2)
            elif handle == 1:
                self.set_rect(x1, y1, nx, ny)
            elif handle in (2, -1):
                delta = (nx, ny) - self.points.mean(axis=0)
                self.points += delta

    def __reduce__(self) -> tuple:
        """Reduce object to picklable state"""
        state = (self.shapeparam, self.points, self.z())
        return (self.__class__, (), state)

    def __setstate__(self, state: tuple) -> None:
        """Set object state from pickled state"""
        param, points, z = state
        # ----------------------------------------------------------------------
        # compatibility with previous version of SegmentShape:
        x1, y1, x2, y2, x3, y3 = points.ravel()
        v12 = np.array((x2 - x1, y2 - y1))
        v13 = np.array((x3 - x1, y3 - y1))
        if np.linalg.norm(v12) < np.linalg.norm(v13):
            # old pickle format
            points = np.flipud(np.roll(points, -1, axis=0))
        # ----------------------------------------------------------------------
        self.points = points
        self.setZ(z)
        self.shapeparam = param
        self.shapeparam.update_item(self)


assert_interfaces_valid(SegmentShape)
